import unittest

from simc_support.game_data.WowSpec import WOWSPECS, get_wow_spec
from simc_support.game_data.Talent import Talent
from simc_support.game_data.Talent import TALENTS
from simc_support.game_data.Talent import get_talent_dict


class TestTalent(unittest.TestCase):
    def test_type(self):
        self.assertTrue(isinstance(TALENTS, list))

    def test_non_emptyness(self):
        self.assertTrue(len(TALENTS) > 0)

    def test_content_types(self):
        for talent in TALENTS:
            with self.subTest(talent=talent):
                self.assertTrue(isinstance(talent, Talent))

    def test_talent_dicts(self):
        for wow_spec in WOWSPECS:
            with self.subTest(wow_spec=wow_spec):
                self.assertTrue(isinstance(get_talent_dict(wow_spec), dict))

    def test_existance_of_all(self):
        for wow_spec in WOWSPECS:
            talents = wow_spec.talents
            for row in range(1, 8):
                for column in range(1, 4):
                    with self.subTest(
                        wow_spec=wow_spec, talents=talents, row=row, column=column
                    ):
                        self.assertTrue(
                            row in talents,
                            f"Talent row {row} for {wow_spec} was not found.",
                        )
                        self.assertTrue(
                            column in talents[row],
                            f"Talent {column} in row {row} for {wow_spec} was not found.",
                        )

    def test_talent_combination_generator(self):
        elemental = get_wow_spec("shaman", "elemental")
        for wow_spec in WOWSPECS:
            combinations = wow_spec.get_talent_combinations()
            with self.subTest(wow_spec=wow_spec):
                self.assertGreater(len(combinations), 0)

                if wow_spec == elemental:
                    self.assertEqual(len(combinations), 243)



"""Exports trinkets from simc_support to three copy= .simc files for SimulationCraft testing purpose.
"""
import typing

from simc_support.game_data.WowSpec import get_wow_spec
from simc_support.game_data.Trinket import get_trinkets_for_spec, Trinket


def print_simc_file(trinket_list: typing.List[Trinket], name: str):
    with open("trinkets_shadowlands_" + name + ".simc", "w") as f:
        f.write("# PROFILE FOR TESTING ONLY!\n")
        f.write(
            "# This file provides all available trinkets for {} and neutral ones.\n".format(
                name.upper()
            )
        )
        f.write(
            "# Use this file to verify whether all trinkets are functioning as expected after changes.\n"
        )
        f.write("# No appropriate drop itemlevel required.\n\n\n")

        for trinket in trinket_list:
            f.write('copy="{}"\n'.format(trinket.name))
            f.write(
                f"trinket1=,id={trinket.item_id},ilevel={trinket.min_itemlevel}\n\n"
            )


def main():

    trinkets = get_trinkets_for_spec(get_wow_spec("warrior", "arms"))
    print_simc_file(trinkets, "str")

    trinkets = get_trinkets_for_spec(get_wow_spec("hunter", "marksmanship"))
    print_simc_file(trinkets, "agi")

    trinkets = get_trinkets_for_spec(get_wow_spec("shaman", "elemental"))
    print_simc_file(trinkets, "int")


if __name__ == "__main__":
    main()



"""
This script updates data_files by using SimulationCrafts casc and dbc extractors.
Basically; execute this script once per patch.

e.g. python self_update.py -s ../../simulationcraft/ --beta --no-load --no-extract
"""
import argparse
import json
import logging
import os
import pathlib
import subprocess
import typing

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)

fh = logging.FileHandler("debug.log", encoding="utf-8")
fh.setLevel(logging.DEBUG)
fh.setFormatter(formatter)

logger.addHandler(ch)
logger.addHandler(fh)

_LOCALES = [
    "en_US",
    "ko_KR",
    "fr_FR",
    "de_DE",
    "zh_CN",
    "es_ES",
    "ru_RU",
    "it_IT",
    "pt_PT",
]

DATA_PATH = "simc_support/game_data/data_files"


def get_spell_names(spells: typing.List[dict], spell_id: int) -> dict:
    for spell in spells:
        if spell_id == spell["id"]:
            spell_names = {"name": spell["name"]}
            for locale in _LOCALES:
                spell_names[f"name_{locale}"] = spell[f"name_{locale}"]
            return spell_names
    raise ValueError(f"No spell with id '{spell_id}' found.")


def merge_information(
    input_dict: dict,
    *,
    filter_function=any,
    match_field="id",
    translation_field="name",
) -> typing.List[dict]:

    logger.debug("Merging")
    result = []
    for i, locale in enumerate(_LOCALES):
        logger.debug(f"  {locale}")
        # prepare
        if i == 0:
            information: dict
            for information in input_dict[locale]:
                if not filter_function(information):
                    continue

                if translation_field:
                    information[f"{translation_field}_{locale}"] = information[
                        translation_field
                    ]

                result.append(information)

        # enrich
        else:
            if not translation_field:
                continue
            for information in result:

                for new_information in input_dict[locale]:
                    if information[match_field] == new_information[match_field]:
                        information[f"{translation_field}_{locale}"] = new_information[
                            translation_field
                        ]

    return result


def handle_arguments() -> argparse.ArgumentParser:
    """Scan user provided arguments and provide the corresponding argparse ArgumentParser.

    Returns:
        argparse.ArgumentParser: [description]
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-s",
        "--simc",
        nargs="?",
        default=None,
        help="Absolute path to your local SimulationCraft version.",
    )
    parser.add_argument(
        "-o",
        "--output",
        nargs="?",
        default=os.path.join(pathlib.Path(__file__).parent.absolute(), "tmp"),
        help="Absolute path to the save location of DB files.",
    )
    parser.add_argument(
        "-w",
        "--wow",
        nargs="?",
        default=None,
        help="Absolute path to your World of Warcaft installation to get hotfixes. E.g. /games/World\ of\ Warcraft/_beta_",
    )
    parser.add_argument(
        "--no-load",
        action="store_true",
        help="Flag disables download from CDN. Instead only already present local files are used.",
    )
    parser.add_argument(
        "--no-extract",
        action="store_true",
        help="Flag disables exrating from db2 files. Instead only already present local files are used.",
    )
    parser.add_argument(
        "--ptr",
        action="store_true",
        help="Download PTR files",
    )
    parser.add_argument(
        "--beta",
        action="store_true",
        help="Download BETA files",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug output",
    )
    parser.add_argument(
        "--env",
        nargs="?",
        default=None,
        help="Absolute path to the executable python-env file.",
    )
    return parser


def get_python(env: str = None) -> str:
    """Find and return the appropriate commandline input to use Python 3+

    Returns:
        str: Either "python" or "python3"
    """

    if env:
        return env

    def is_python() -> bool:
        cmd = ["python", "--version"]
        logger.debug(cmd)
        output = subprocess.run(cmd, stdout=subprocess.PIPE)
        return output.stdout.split()[1].startswith(b"3")

    def is_python3() -> bool:
        cmd = ["python3", "--version"]
        logger.debug(cmd)
        output = subprocess.run(cmd, stdout=subprocess.PIPE)
        return output.stdout.split()[1].startswith(b"3")

    if not (is_python() or is_python3()):
        raise SystemError("Python 3 is required.")

    return "python" if is_python() else "python3"


def casc(args: object) -> None:
    """Download all db2 files of all locales described in _LOCALES.

    Args:
        args (object): [description]

    Raises:
        SystemError: If Python 3+ is not available in commandline neither via 'python' nor 'python3'.
    """
    if args.no_load:
        logger.info("Skipping download")
        return
    logger.info("Downloading files (casc)")

    python = get_python(args.env)

    command = [
        python,
        "casc_extract.py",
        "--cdn",
        "-m",
        "batch",
    ]
    if args.ptr:
        command.append("--ptr")
    elif args.beta:
        command.append("--beta")

    for locale in _LOCALES:
        logger.info(locale)
        full_command = command + [
            "--locale",
            locale,
            "-o",
            os.path.join(args.output, locale),
        ]
        logger.debug(full_command)
        process = subprocess.Popen(
            full_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=os.path.join(args.simc, "casc_extract"),
        )
        # ['python', 'casc_extract.py', '--cdn', '-m', 'batch', '--locale', 'pt_PT', '-o', 'D:\\Programme\\simc_support\\simc_support\\tmp\\pt_PT']
        while True:
            output = process.stdout.readline()
            if output == b"" and process.poll() is not None:
                break
            if output:
                logger.debug(output.strip().decode("ascii"))
        rc = process.poll()
        if rc != 0:
            logger.error(
                f"Error occured while loading {locale}. {process.communicate()[1]}"
            )

    logger.info("Downloaded files")


def get_compiled_data_path(args: object, locale: str) -> str:
    return os.path.join(args.output, "compiled", locale)


def dbc(args: object, files: typing.List[str]) -> None:
    """Extract and transforms files into JSON format.

    Args:
        args (object): [description]
        files (typing.List[str]): [description]
    """
    if args.no_extract:
        logger.info("Skipping extraction")
        return
    logger.info("Extracting files (dbc)")

    python = get_python(args.env)

    wow_versions = os.listdir(os.path.join(args.output, _LOCALES[0]))
    logger.debug(f" Detected wow versions: {wow_versions}")
    wow_version = wow_versions[-1]

    command = [
        python,
        "dbc_extract.py",
        "-b",
        wow_version,
        "-t",
        "json",
    ]
    if args.wow:
        command.append("--hotfix")
        command.append(os.path.join(args.wow, "Cache/ADB/enUS/DBCache.bin"))

    for file in files:
        for locale in _LOCALES:
            logger.info(f"{file} {locale}")

            cmd = command + [
                "-p",
                os.path.join(args.output, locale, wow_version, "DBFilesClient"),
                file,
            ]

            pathlib.Path(os.path.join(get_compiled_data_path(args, locale))).mkdir(
                parents=True, exist_ok=True
            )

            logger.debug(cmd)
            process = subprocess.run(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=os.path.join(args.simc, "dbc_extract3"),
            )
            if process.returncode != 0:
                logger.error(
                    f"Error occured while extracting {file} {locale}. {process.stderr.decode()}"
                )
            else:
                with open(
                    os.path.join(get_compiled_data_path(args, locale), f"{file}.json"),
                    "wb",
                ) as f:
                    f.write(process.stdout)


def update_trinkets(args: object) -> None:
    """Update data_files/trinkets.json

    Args:
        args (object): [description]

    Returns:
        None
    """
    logger.info("Update trinkets")

    ITEMSPARSE = "ItemSparse"
    ITEMXITEMEFFECT = "ItemXItemEffect"
    ITEMEFFECT = "ItemEffect"
    JOURNAL_ENCOUNTER_ITEM = "JournalEncounterItem"
    JOURNAL_ENCOUNTER = "JournalEncounter"
    JOURNAL_INSTANCE = "JournalInstance"
    MAP = "Map"

    WHITELIST = [
        181457,
        184842,
    ]

    def is_trinket(item: dict) -> bool:
        """See https://github.com/simulationcraft/simc/blob/shadowlands/engine/dbc/data_enums.hh#L335"""
        return item.get("inv_type") == 12

    def is_gte_rare(item: dict) -> bool:
        """See https://github.com/simulationcraft/simc/blob/shadowlands/engine/dbc/data_enums.hh#L369"""
        return item.get("quality") >= 3

    def is_shadowlands(item: dict) -> bool:
        """Tests for itemlevel and chracter level requirement."""

        def is_expansion(item: dict) -> bool:
            """Checks expansion id."""
            return item["id_expansion"] == 9

        def is_gte_level(item: dict) -> bool:
            """An Expansion starts at the last possible character level of the previous one."""
            return item.get("req_level") >= 50

        def is_gte_ilevel(item: dict) -> bool:
            """Value of normal Dungeon items at level 50."""
            return item.get("ilevel") >= 148

        # Unbound Changeling is somehow available for lvl 1
        return is_expansion(item) or is_gte_ilevel(item)  # and is_gte_level(item)

    def is_whitelisted(item: dict) -> bool:
        return item.get("id") in WHITELIST or item.get("name") in WHITELIST

    def is_approved(item: dict) -> bool:
        return (
            is_trinket(item)
            and is_gte_rare(item)
            and is_shadowlands(item)
            or is_whitelisted(item)
        )

    dbc(
        args,
        [
            ITEMEFFECT,
            ITEMSPARSE,
            ITEMXITEMEFFECT,
            JOURNAL_ENCOUNTER,
            JOURNAL_ENCOUNTER_ITEM,
            JOURNAL_INSTANCE,
            MAP,
        ],
    )

    data = {}

    for locale in _LOCALES:
        with open(
            os.path.join(get_compiled_data_path(args, locale), f"{ITEMSPARSE}.json"),
            "r",
        ) as f:
            items = json.load(f)

        data[locale] = [item for item in items if is_approved(item)]

    # load itemeffect table
    with open(
        os.path.join(get_compiled_data_path(args, _LOCALES[0]), f"{ITEMEFFECT}.json"),
        "r",
    ) as f:
        _itemeffects = json.load(f)

    itemeffects = {}
    for effect in _itemeffects:
        itemeffects[effect["id"]] = effect

    logger.debug(itemeffects)

    # load ItemXItemEffect table
    with open(
        os.path.join(
            get_compiled_data_path(args, _LOCALES[0]), f"{ITEMXITEMEFFECT}.json"
        ),
        "r",
    ) as f:
        _itemxitemeffects = json.load(f)

    itemxitemeffects = {}
    for effect in _itemxitemeffects:
        if effect["id_parent"] in itemxitemeffects:
            itemxitemeffects[effect["id_parent"]].append(
                itemeffects[effect["id_item_effect"]]
            )
        else:
            itemxitemeffects[effect["id_parent"]] = [
                itemeffects[effect["id_item_effect"]]
            ]

    def is_on_use(item_id: int) -> bool:
        if item_id in itemxitemeffects:
            return any(
                map(lambda row: row["trigger_type"] == 0, itemxitemeffects[item_id])
            )
        return False

    def get_spec_mask(item_id: int) -> list:
        """Turns out...trinkets add only 0 as specialization ids...worthless approach.

        Args:
            item_id (int): [description]

        Returns:
            list: [description]
        """
        mask = []
        for effect in itemeffects:
            if item_id == effect["id_parent"]:
                mask.append(effect["id_specialization"])
        if len(mask) == 0:
            logger.info(f"No specializations found for {item_id}")
        else:
            logger.info(f"Specializations found for {item_id}: {mask}")
        return mask

    # load encounter_items table
    with open(
        os.path.join(
            get_compiled_data_path(args, _LOCALES[0]), f"{JOURNAL_ENCOUNTER_ITEM}.json"
        ),
        "r",
    ) as f:
        encounter_items = json.load(f)

    def get_id_encounter(item_id: int) -> typing.Optional[int]:
        for item in encounter_items:
            if item["id_item"] == item_id:
                return item["id_encounter"]
        return None

    # load encounter table
    with open(
        os.path.join(
            get_compiled_data_path(args, _LOCALES[0]), f"{JOURNAL_ENCOUNTER}.json"
        ),
        "r",
    ) as f:
        encounters = json.load(f)

    def id_journal_instance(id_encounter: int) -> typing.Optional[int]:
        if not id_encounter:
            return None
        for encounter in encounters:
            if encounter["id"] == id_encounter:
                return encounter["id_journal_instance"]
        return None

    # load instance table
    with open(
        os.path.join(
            get_compiled_data_path(args, _LOCALES[0]), f"{JOURNAL_INSTANCE}.json"
        ),
        "r",
    ) as f:
        instances = json.load(f)

    def get_id_map(id_journal_instance: int) -> typing.Optional[int]:
        if not id_journal_instance:
            return None
        for instance in instances:
            if instance["id"] == id_journal_instance:
                return instance["map"]
        return None

    # load map table
    with open(
        os.path.join(get_compiled_data_path(args, _LOCALES[0]), f"{MAP}.json"),
        "r",
    ) as f:
        maps = json.load(f)

    def get_instance_type(id_map: int) -> typing.Optional[int]:
        if not id_map:
            return None
        for map in maps:
            if map["id"] == id_map:
                return map["instance_type"]
        return None

    trinkets = merge_information(data, filter_function=is_approved)
    for trinket in trinkets:
        trinket["on_use"] = is_on_use(trinket["id"])
        trinket["id_encounter"] = get_id_encounter(trinket["id"])
        trinket["id_journal_instance"] = id_journal_instance(trinket["id_encounter"])
        trinket["id_map"] = get_id_map(trinket["id_journal_instance"])
        trinket["instance_type"] = get_instance_type(trinket["id_map"])

    logger.debug(trinkets)

    with open(os.path.join(DATA_PATH, "trinkets.json"), "w", encoding="utf-8") as f:
        json.dump(trinkets, f, ensure_ascii=False)

    logger.info(f"Updated {len(trinkets)} trinkets")


def update_wow_classes(args: object) -> None:
    logger.info("Update WowClasses")

    # TODO: class - race connection can be established via ChrClassRaceSex table
    CHRCLASSES = "ChrClasses"
    dbc(
        args,
        [
            CHRCLASSES,
        ],
    )
    data = {}

    for locale in _LOCALES:
        with open(
            os.path.join(get_compiled_data_path(args, locale), f"{CHRCLASSES}.json"),
            "r",
        ) as f:
            data[locale] = json.load(f)

    wow_classes = []

    wow_classes = merge_information(data, translation_field="name_lang")

    with open(os.path.join(DATA_PATH, "wow_classes.json"), "w", encoding="utf-8") as f:
        json.dump(wow_classes, f, ensure_ascii=False)

    logger.info(f"Updated {len(wow_classes)} WowClasses")


def update_covenants(args: object) -> None:
    logger.info("Update covenants")

    COVENANTS = "Covenant"

    dbc(args, [COVENANTS])
    data = {}
    for locale in _LOCALES:
        with open(
            os.path.join(get_compiled_data_path(args, locale), f"{COVENANTS}.json"),
            "r",
        ) as f:
            data[locale] = json.load(f)

    covenants = merge_information(data)

    with open(os.path.join(DATA_PATH, "covenants.json"), "w", encoding="utf-8") as f:
        json.dump(covenants, f, ensure_ascii=False)

    logger.info(f"Updated {len(covenants)} covenants")


# SoulbindConduit, SoulbindConduitItem, SoulbindConduitRank for the conduit stuff


def update_soul_binds(args: object) -> None:
    logger.info("Update soul binds")

    SOULBINDS = "Soulbind"
    TALENTS = "GarrTalent"
    RANKS = "GarrTalentRank"
    SPELLS = "SpellName"

    dbc(
        args,
        [
            SOULBINDS,
            TALENTS,
            RANKS,
            SPELLS,
        ],
    )
    data = {SOULBINDS: {}, TALENTS: {}, RANKS: {}, SPELLS: {}}
    for key in data:
        for locale in _LOCALES:
            with open(
                os.path.join(get_compiled_data_path(args, locale), f"{key}.json"),
                "r",
            ) as f:
                data[key][locale] = json.load(f)

    soul_binds = merge_information(data[SOULBINDS])

    talents = merge_information(data[TALENTS])

    ranks = merge_information(data[RANKS], translation_field=None)

    def get_spell_id(talent: dict) -> int:
        for rank in ranks:
            if rank["id_parent"] == talent["id"]:
                return rank["id_spell"]

    logger.debug("Enriching talent information")
    # enrich with talents
    talent_spells = []
    for soul_bind in soul_binds:
        logger.debug(f"Soul Bind {soul_bind['id']}")
        soul_bind["talents"] = [
            talent
            for talent in talents
            if talent["id_parent"] == soul_bind["id_garr_talent_tree"]
        ]
        for talent in soul_bind["talents"]:
            logger.debug(f"  Talent {talent['id']}")
            talent["spell_id"] = get_spell_id(talent)
            talent_spells.append(talent["spell_id"])

    # enrich talents with spells and spell translations
    def is_relevant_spell(information: dict) -> bool:
        return information["id"] in talent_spells

    spells = merge_information(data[SPELLS], filter_function=is_relevant_spell)

    for soul_bind in soul_binds:
        logger.debug(f"Soul Bind {soul_bind['id']}")
        for talent in soul_bind["talents"]:
            if talent["spell_id"] == 0:
                continue
            # node names are overwritten by their spells
            spell_names = get_spell_names(spells, talent["spell_id"])
            for key in spell_names:
                talent[key] = spell_names[key]

    with open(os.path.join(DATA_PATH, "soul_binds.json"), "w", encoding="utf-8") as f:
        json.dump(soul_binds, f, ensure_ascii=False)

    logger.info(f"Updated {len(soul_binds)} soul binds")


def update_legendaries(args: object) -> None:
    logger.info("Update legendaries")

    RUNEFORGELEGENDARYABILITY = "RuneforgeLegendaryAbility"
    SPECSETMEMBER = "SpecSetMember"

    dbc(
        args,
        [
            RUNEFORGELEGENDARYABILITY,
            SPECSETMEMBER,
        ],
    )
    data = {
        RUNEFORGELEGENDARYABILITY: {},
        SPECSETMEMBER: {},
    }
    for key in data:
        for locale in _LOCALES:
            with open(
                os.path.join(get_compiled_data_path(args, locale), f"{key}.json"),
                "r",
            ) as f:
                data[key][locale] = json.load(f)

    legendaries = merge_information(data[RUNEFORGELEGENDARYABILITY])
    specs = data[SPECSETMEMBER][_LOCALES[0]]

    for legendary in legendaries:
        legendary["spec_ids"] = list(
            [
                spec["id_spec"]
                for spec in specs
                if spec["id_parent"] == legendary["id_spec_set"]
            ]
        )

    with open(os.path.join(DATA_PATH, "legendaries.json"), "w", encoding="utf-8") as f:
        json.dump(legendaries, f, ensure_ascii=False)

    logger.info(f"Updated {len(legendaries)} legendaries")


def update_conduits(args: object) -> None:
    logger.info("Update conduits")

    CONDUIT = "SoulbindConduit"
    # ITEM = "SoulbindConduitItem"
    RANK = "SoulbindConduitRank"
    SPECSETMEMBER = "SpecSetMember"
    SPELLS = "SpellName"

    WHITELIST = []

    dbc(
        args,
        [
            CONDUIT,
            RANK,
            SPECSETMEMBER,
            SPELLS,
        ],
    )
    data = {
        CONDUIT: {},
        RANK: {},
        SPECSETMEMBER: {},
        SPELLS: {},
    }
    for key in data:
        for locale in _LOCALES:
            with open(
                os.path.join(get_compiled_data_path(args, locale), f"{key}.json"),
                "r",
            ) as f:
                data[key][locale] = json.load(f)

    conduits = merge_information(data[CONDUIT], translation_field=None)
    ranks = merge_information(data[RANK], translation_field=None)
    specs = data[SPECSETMEMBER][_LOCALES[0]]

    def _get_spell_id(conduit: dict) -> typing.Optional[int]:
        for rank in ranks:
            if rank["id_parent"] == conduit["id"]:
                return rank["id_spell"]
        return None

    def _get_ranks(conduit: dict) -> typing.List[int]:
        return list(
            [rank["rank"] for rank in ranks if rank["id_parent"] == conduit["id"]]
        )

    spell_ids = []
    for conduit in conduits:
        conduit["spec_ids"] = list(
            [
                spec["id_spec"]
                for spec in specs
                if spec["id_parent"] == conduit["id_spec_set"]
            ]
        )
        conduit["spell_id"] = _get_spell_id(conduit)
        conduit["ranks"] = sorted(_get_ranks(conduit))
        spell_ids.append(conduit["spell_id"])

    # enrich talents with spells and spell translations
    def is_relevant_spell(spell: dict) -> bool:
        return spell["id"] in spell_ids

    spells = merge_information(data[SPELLS], filter_function=is_relevant_spell)

    for conduit in conduits:
        try:
            translations = get_spell_names(spells, conduit["spell_id"])
        except ValueError:
            translations = {"name": None}
            for locale in _LOCALES:
                translations[f"name_{locale}"] = None
        for key in translations:
            conduit[key] = translations[key]

    def _is_conduit(conduit: dict) -> bool:
        return (
            conduit["name"]
            and len(conduit["ranks"]) == 15
            and len(conduit["spec_ids"]) > 0
            or conduit["id"] in WHITELIST
        )

    conduits = list([conduit for conduit in conduits if _is_conduit(conduit)])

    with open(os.path.join(DATA_PATH, "conduits.json"), "w", encoding="utf-8") as f:
        json.dump(conduits, f, ensure_ascii=False)

    logger.info(f"Updated {len(conduits)} conduits")


def update_talents(args: object) -> None:
    logger.info("Update talents")

    TALENTS = "Talent"
    SPELLS = "SpellName"

    dbc(args, [TALENTS, SPELLS])
    data = {
        TALENTS: {},
        SPELLS: {},
    }
    for key in data:
        for locale in _LOCALES:
            with open(
                os.path.join(get_compiled_data_path(args, locale), f"{key}.json"),
                "r",
            ) as f:
                data[key][locale] = json.load(f)

    talents = data[TALENTS][_LOCALES[0]]

    spell_ids = list([talent["id_spell"] for talent in talents])
    spells = merge_information(
        data[SPELLS], filter_function=lambda spell: spell["id"] in spell_ids
    )

    for talent in talents:
        spell_names = get_spell_names(spells, talent["id_spell"])
        for key in spell_names:
            talent[key] = spell_names[key]

    with open(os.path.join(DATA_PATH, "talents.json"), "w", encoding="utf-8") as f:
        json.dump(talents, f, ensure_ascii=False)

    logger.info(f"Updated {len(talents)} talents")


def update_shards_of_dominations(args: object) -> None:
    logger.info("Update Shards of Dominations")

    ITEM_SPARSE = "ItemSparse"
    GEM_PROPERTIES = "GemProperties"
    SPELL_ITEM_ENCHANTMENT = "SpellItemEnchantment"
    SPELL_NAME = "SpellName"
    SPELL_MISC = "SpellMisc"

    dbc(
        args,
        [
            ITEM_SPARSE,
            GEM_PROPERTIES,
            SPELL_ITEM_ENCHANTMENT,
            SPELL_NAME,
            SPELL_MISC,
        ],
    )
    data = {
        ITEM_SPARSE: {},
        GEM_PROPERTIES: {},
        SPELL_ITEM_ENCHANTMENT: {},
        SPELL_NAME: {},
        SPELL_MISC: {},
    }

    for key in data:
        for locale in _LOCALES:
            with open(
                os.path.join(get_compiled_data_path(args, locale), f"{key}.json"),
                "r",
            ) as f:
                data[key][locale] = json.load(f)

    def is_shard(entry: dict) -> bool:
        not_a_shard = [
            187120,
            186773,
        ]
        if entry["id"] in not_a_shard:
            return False
        return (
            entry["id_expansion"] == 7
            and entry["ilevel"] >= 255
            and entry["quality"] == 4
            and entry["req_level"] >= 50
            and entry["gem_props"] != 0
        )

    shards = list(filter(is_shard, data[ITEM_SPARSE]["en_US"]))

    shard_ids = list(set([shard["gem_props"] for shard in shards]))
    logger.info(f"shard_ids ({len(shard_ids)}): {shard_ids}")

    def is_relevant_gem_prop(entry: dict) -> bool:
        return entry["id"] in shard_ids

    relevant_gems = list(filter(is_relevant_gem_prop, data[GEM_PROPERTIES]["en_US"]))

    enchant_ids = [gem_prop["id_enchant"] for gem_prop in relevant_gems]
    logger.info(f"enchant_ids ({len(enchant_ids)}): {enchant_ids}")

    def is_shard_enchantment(entry: dict) -> bool:
        return entry["id"] in enchant_ids

    # gem_properties = data[GEM_PROPERTIES]
    spell_item_enchantments = list(
        filter(is_shard_enchantment, data[SPELL_ITEM_ENCHANTMENT]["en_US"])
    )

    properties = ("id_property_1", "id_property_2", "id_property_3")

    cutoff = 0
    # logger.info(json.dumps(spell_item_enchantments))
    shard_enchantment_ids = []
    for prop in properties:
        shard_enchantment_ids += [
            enchant[prop]
            for enchant in spell_item_enchantments
            if enchant[prop] > cutoff
        ]
    shard_enchantment_ids = list(set(shard_enchantment_ids))
    logger.info(
        f"shard_enchantment_ids ({len(shard_enchantment_ids)}): {shard_enchantment_ids}"
    )

    def is_shard_spell(entry: dict) -> bool:
        return entry["id"] in shard_enchantment_ids

    spell_names = merge_information(data[SPELL_NAME], filter_function=is_shard_spell)

    names = [spell["name"] for spell in spell_names]
    logger.info(f"names ({len(names)}): {names}")

    def is_misc(entry: dict) -> dict:
        return entry["id_parent"] in shard_enchantment_ids

    spell_miscs = list(filter(is_misc, data[SPELL_MISC]["en_US"]))

    for shard in shards:
        for gem in relevant_gems:
            if shard["gem_props"] == gem["id"]:
                shard["id_enchant"] = gem["id_enchant"]
        for enchant in spell_item_enchantments:
            if shard["id_enchant"] == enchant["id"]:
                for prop in properties:
                    shard[prop] = enchant[prop]
        for name in spell_names:
            if name["id"] in [shard[prop] for prop in properties]:
                for key in name.keys():
                    if "name_" in key:
                        shard[key] = name[key]
        for misc in spell_miscs:
            if misc["id_parent"] in [shard[prop] for prop in properties]:
                shard["school"] = misc["school"]
        # cleanse unwanted keys
        unwanted_keys = [
            "desc",
            "pad2",
            "pad1",
            "pad0",
            "dmg_range",
            "item_limit_category",
            "duration",
            "item_damage_modifier",
            "bag_family",
            "ranged_mod_range",
            "stackable",
            "max_count",
            "id_content_tuning",
            "sell_price",
            "buy_price",
            "unk_3",
            "unk_2",
            "unk_1",
            "flags_1",
            "flags_2",
            "flags_3",
            "flags_4",
            "faction_conv_id",
            "unk_901_1",
            "req_spell",
            "id_curve",
            # "id_name_desc",
            "unk_l72_1",
            "id_holiday",
            "socket_bonus",
            "totem_category",
            "map",
            "area_1",
            "area_2",
            "item_set",
            "id_lock",
            "start_quest",
            "page_text",
            "delay",
            "req_rep_faction",
            "req_skill_rank",
            "req_skill",
            "id_artifact",
            "unk_6",
            "unk_7",
            "socket_color_1",
            "socket_color_2",
            "socket_color_3",
            "sheath",
            "page_mat",
            "id_lang",
            "bonding",
            "damage_type",
            "container_slots",
            "req_rep_rank",
            "unk_5",
            "unk_4",
            "inv_type",
        ]
        for i in range(1, 11):
            unwanted_keys += [
                f"stat_socket_mul_{i}",
                f"stat_alloc_{i}",
                f"stat_type_{i}",
            ]
        for key in unwanted_keys:
            del shard[key]

    # save data to file
    with open(
        os.path.join(DATA_PATH, "damnation_shards.json"), "w", encoding="utf-8"
    ) as f:
        json.dump(shards, f, ensure_ascii=False)

    logger.info(f"Updated {len(shards)} DamnationShards")


def main() -> None:
    args = handle_arguments().parse_args()
    if args.debug:
        logger.setLevel(logging.DEBUG)
    casc(args)
    update_trinkets(args)
    update_wow_classes(args)
    update_covenants(args)
    update_soul_binds(args)
    update_legendaries(args)
    update_conduits(args)
    update_talents(args)
    update_shards_of_dominations(args)
    # anima powers? https://ptr.wowhead.com/news/how-the-tormented-seasonal-mythic-affix-works-get-anima-powers-from-lieutenants-322665

    logger.debug("self_update done")


if __name__ == "__main__":
    main()



from simc_support.game_data.SimcObject import SimcObject


class Faction(SimcObject):
    """Factions divide the player base, usually have differences in
    storylines, and quests.
    """

    @property
    def name(self) -> str:
        """Return full name of the Faction.

        Returns:
            str: Name of the Faction
        """
        return self.full_name


ALLIANCE = Faction("Alliance", 'alliance')
HORDE = Faction("Horde", 'horde')

FACTIONS = (ALLIANCE, HORDE)



import json
import pkg_resources
import typing


from simc_support.game_data.Language import _get_translations
from simc_support.game_data.Language import Translation
from simc_support.game_data.SimcObject import SimcObject


class Covenant(SimcObject):
    """Shadowlands specific faction. Each character joins one."""

    def __init__(
        self,
        *args,
        id: int,
        translations: typing.Union[typing.Dict, Translation],
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.id = id

        if isinstance(translations, Translation):
            self.translations = translations
        elif isinstance(translations, dict):
            self.translations = Translation(translations=translations)
        else:
            raise TypeError(
                "translations must either be a dictionary or a Translaton object."
            )


def _load_covenants() -> typing.List[Covenant]:
    with pkg_resources.resource_stream(
        __name__, "/".join(("data_files", "covenants.json"))
    ) as f:
        loaded_covenants = json.load(f)

    covenants = []
    for covenant in loaded_covenants:
        covenants.append(
            Covenant(
                id=covenant["id"],
                translations=_get_translations(covenant),
                full_name=covenant["name"],
                simc_name=covenant["name"].lower().replace(" ", "_"),
            )
        )
    return covenants


COVENANTS = _load_covenants()


def get_covenant(*, id: int = None, name: str = None) -> Covenant:
    if not id and not name:
        raise ValueError("Function requires either 'id' or 'name'.")

    for covenant in COVENANTS:
        if (
            name
            and name.lower().replace(" ", "_") == covenant.simc_name
            or id
            and id == covenant.id
        ):
            return covenant
    raise ValueError(f"No Covenant with name '{name}' found.")



# these values are used throughout the code to determine itemlevel "borders" of items
# usually used as a catch up mechanic, drops usually all items from previous content
TRADER_TOKEN = -1
WORLD_QUEST_ITEMLEVELS = [
    144,
    148,
    161,
    171,
    200,
]  # guess
CALLINGS = WORLD_QUEST_ITEMLEVELS

DUNGEON_NORMAL = 184
DUNGEON_HEROIC = 197
DUNGEON_MYTHIC = 210
DUNGEON_MYTHIC_LEVELS = [
    # 187,
    # 190,
    # 194,
    # 197,
    # 200,
    # 203,  # weekly
    # 204,
    # 207,
    210,
    213,
    216,
    220,
    223,
    226,  # + weekly
    229,  # + weekly
    233,  # + weekly
    236,  # + weekly
    239,  # weekly
    242,  # weekly
    246,  # weekly
    249,  # weekly
    252,  # weekly
]
DUNGEON = [DUNGEON_NORMAL] + [DUNGEON_HEROIC] + [DUNGEON_MYTHIC] + DUNGEON_MYTHIC_LEVELS
TAZAVESH = 226

# raids
CASTLE_NATHRIA = [187, 200, 213, 226]
CASTLE_NATHRIA_ENDBOSSES = list([ilvl + 7 for ilvl in CASTLE_NATHRIA])

SANCTUM_OF_DOMINATION = [213, 226, 239, 252]
SANCTUM_OF_DOMINATION_ENDBOSSES = list([ilvl + 7 for ilvl in SANCTUM_OF_DOMINATION])

UPPER_BOUND = CASTLE_NATHRIA_ENDBOSSES[-1]

PVP_HONOR = [
    # 158,
    # 164,
    # 171,
    177,
    184,
    190,
    197,
    203,  # 9.1
    210,  # 9.1
    216,  # 9.1
]
PVP_CONQUEST = [
    # 200,
    # 207,
    # 213,
    220,
    226,
    233,  # 9.1
    240,  # 9.1
    246,  # 9.1
]
PVP = PVP_HONOR + PVP_CONQUEST


RARE_MOB = {
    155: [
        155,
        164,
    ]
}



import json
import pkg_resources
import typing

try:
    from simc_support.game_data.WowSpec import WowSpec
except ImportError:
    WowSpec: object = None  # type: ignore # test setup
from simc_support.game_data.Language import Translation, _get_translations
from simc_support.game_data.SimcObject import SimcObject


class Talent(SimcObject):
    def __init__(
        self,
        wow_class_id: int,
        wow_spec_id: int,
        name: str,
        spell_id: int,
        row: int,
        column: int,
        translations: Translation,
    ):
        super().__init__(full_name=name)
        self.wow_class_id = int(wow_class_id)
        self.wow_spec_id = int(wow_spec_id)
        self.name = str(name)
        self.spell_id = int(spell_id)
        self.row = int(row)
        self.column = int(column)
        if isinstance(translations, Translation):
            self.translations = translations
        else:
            self.translations = Translation(translations=translations)

    def get_dict(self) -> dict:
        return {"name": self.name, "spell_id": self.spell_id}

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.simc_name


def _load_talents() -> typing.List[Talent]:
    with pkg_resources.resource_stream(
        __name__, "/".join(("data_files", "talents.json"))
    ) as f:
        loaded_talents = json.load(f)

    talents = []
    for talent in loaded_talents:
        talents.append(
            Talent(
                wow_class_id=talent["class_id"],
                wow_spec_id=talent["spec_id"],
                name=talent["name"],
                spell_id=talent["id_spell"],
                row=talent["row"],
                column=talent["col"],
                translations=_get_translations(talent),
            )
        )

    return talents


TALENTS = _load_talents()


def get_talent_dict(wow_spec: WowSpec, ptr: bool = None) -> dict:
    """Return the dict of all talents available to this spec. Structur: row -> column -> name, spell_id

    Arguments:
        wow_spec {WowSpec} -- [description]

    Returns:
        dict -- row -> column -> name, spell_id
    """
    talents = list(
        [
            talent
            for talent in TALENTS
            if talent.wow_spec_id == wow_spec.id
            or talent.wow_spec_id == 0  # shared between specs
            and talent.wow_class_id == wow_spec.wow_class.id
        ]
    )

    result = {}  # type: ignore # newer python versions and thus mypy versions want an annotation here
    for row in range(1, 8):
        result[row] = {}

    for talent in talents:
        # spec specific information precedes class information
        if talent.wow_spec_id == 0 and (talent.column + 1) in result[talent.row + 1]:
            continue
        result[talent.row + 1][talent.column + 1] = talent.get_dict()

    return result



def get_talents_for_spec(wow_spec: WowSpec) -> typing.Tuple[Talent, ...]:
    return tuple(
        [
            talent
            for talent in TALENTS
            if talent.wow_spec_id == 0
            and talent.wow_class_id == wow_spec.wow_class.id
            or talent.wow_spec_id == wow_spec.id
        ]
    )



from enum import Enum
import json
import pkg_resources
import typing

from simc_support.game_data.SimcObject import SimcObject
from simc_support.game_data.Language import _get_translations
from simc_support.game_data.Language import Translation


class ShardType(Enum):
    BLOOD = "blood"
    FROST = "frost"
    UNHOLY = "unholy"

    @staticmethod
    def get_shard_type(school: int) -> "ShardType":
        mapping = {
            16: ShardType.FROST,
            32: ShardType.UNHOLY,
            64: ShardType.BLOOD,
        }
        return mapping[school]


class DominationShard(SimcObject):
    def __init__(
        self,
        *args,
        gem_id: int,
        school: int,
        translations: Translation,
        **kwargs,
    ):
        """Domination Socket of the second Raid of Shadowlands

        Args:
            gem_id (int): id of the gem
            school (int): int of the spell-school
            translations (Translation): name of the item in all languages
        """
        super().__init__(translations.US, *args, **kwargs)
        self.translations = translations
        self.name: str = self.translations.US
        self.gem_id: int = int(gem_id)
        self.school: int = int(school)
        self.school_type = ShardType.get_shard_type(self.school)


def _load_domination_shards() -> typing.List[DominationShard]:
    with pkg_resources.resource_stream(
        __name__, "/".join(("data_files", "domination_shards.json"))
    ) as f:
        loaded_domination_shards = json.load(f)

    domination_shards = []
    for shard in loaded_domination_shards:
        domination_shards.append(
            DominationShard(
                gem_id=shard["id"],
                school=shard["school"],
                translations=_get_translations(shard),
            )
        )
    return domination_shards


DOMINATION_SHARDS: typing.List[DominationShard] = _load_domination_shards()



import unittest

from simc_support.game_data.Conduit import Conduit
from simc_support.game_data.Conduit import CONDUITS
from simc_support.game_data.Conduit import get_conduits_for_spec
from simc_support.game_data.WowSpec import get_wow_spec


class TestConduit(unittest.TestCase):
    def setUp(self) -> None:
        self.spec = get_wow_spec("Shaman", "Elemental")
        self.conduits = get_conduits_for_spec(self.spec)

    def test_type(self):
        self.assertTrue(isinstance(CONDUITS, list))

    def test_non_emptyness(self):
        self.assertTrue(len(CONDUITS) > 0)

    def test_content_types(self):
        for conduit in CONDUITS:
            with self.subTest(conduit=conduit):
                self.assertTrue(isinstance(conduit, Conduit))

    def test_get_conduits_for_spec_existence(self):
        self.assertTrue(len(self.conduits) > 0)

    def test_get_conduits_for_spec_type(self):
        self.assertTrue(isinstance(self.conduits, tuple))

    def test_get_conduits_for_spec_types(self):
        self.assertTrue(
            all([isinstance(conduit, Conduit) for conduit in self.conduits])
        )

    def test_has_finesse(self):
        self.assertTrue(any([conduit.is_finesse for conduit in CONDUITS]))

    def test_has_potency(self):
        self.assertTrue(any([conduit.is_potency for conduit in CONDUITS]))

    def test_has_endurance(self):
        self.assertTrue(any([conduit.is_endurance for conduit in CONDUITS]))



import unittest

from simc_support.game_data.DominationShard import DOMINATION_SHARDS
from simc_support.game_data.DominationShard import ShardType


class TestDominationShards(unittest.TestCase):
    def test_type(self):
        self.assertTrue(isinstance(DOMINATION_SHARDS, list))

    def test_non_emptyness(self):
        self.assertTrue(len(DOMINATION_SHARDS) > 0)

    def test_for_each_shard_type(self):
        """At least one shard for each shard type is available."""
        for type in ShardType:
            with self.subTest(type=type):
                found = False
                for shard in DOMINATION_SHARDS:
                    if type == shard.school_type:
                        found = True
                        break
                self.assertTrue(found)



import enum


class Language(enum.Enum):
    BR = "pt_BR"
    CN = "cn_CN"
    DE = "de_DE"
    ES = "es_ES"
    FR = "fr_FR"
    IT = "it_IT"
    KR = "ko_KR"
    RU = "ru_RU"
    US = "en_US"


class Translation(object):
    """Full translation information. All languages are present."""

    def __init__(
        self,
        pt_BR: str = None,
        cn_CN: str = None,
        de_DE: str = None,
        es_ES: str = None,
        fr_FR: str = None,
        it_IT: str = None,
        ko_KR: str = None,
        ru_RU: str = None,
        en_US: str = None,
        translations: dict = None,
        *args,
        **kwargs,
    ):
        super().__init__()

        languages = [language.name for language in Language]

        if translations:
            if len(set(translations.keys()).intersection(languages)) == len(languages):
                for language in translations:
                    setattr(self, language, translations[language])
            elif isinstance(list(translations)[0], Language):
                if len(
                    set([key.name for key in list(translations)]).intersection(
                        languages
                    )
                ) == len(languages):
                    for language in translations:
                        setattr(self, language.name, translations[language])
            else:
                raise ValueError(
                    f"'translations'-dictionary didn't contain expected keys. Got: {sorted(list(translations))} Expected keys: {sorted(languages)}"
                )
        elif (
            cn_CN
            and de_DE
            and es_ES
            and fr_FR
            and it_IT
            and ko_KR
            and ru_RU
            and en_US
            and pt_BR
        ):
            self.CN = cn_CN
            self.DE = de_DE
            self.ES = es_ES
            self.FR = fr_FR
            self.IT = it_IT
            self.KR = ko_KR
            self.RU = ru_RU
            self.US = en_US
            self.BR = pt_BR
        else:
            raise ValueError(
                "Expected more information. Please make sure you have translations for all languages."
            )

    def get_dict(self) -> dict:
        response = {}
        for lang in Language:
            response[lang.value] = getattr(self, lang.name)

        return response


class EmptyTranslation(Translation):
    """If you don't have translations handy, but need to provide a translations object."""

    def __init__(self):
        languages = {language.name: "" for language in Language}
        super().__init__(self, translations=languages)


def _get_translations(item: dict, unified_key: str = "name") -> Translation:
    keys = [
        "en_US",
        "ko_KR",
        "fr_FR",
        "de_DE",
        "zh_CN",
        "es_ES",
        "ru_RU",
        "it_IT",
        "pt_PT",
    ]
    d = {}
    for key in keys:
        d[key.split("_")[1]] = item[f"{unified_key}_{key}"]

    d["BR"] = d["PT"]
    d.pop("PT")

    return Translation(translations=d)



import itertools
import typing

from simc_support.game_data import Language, WowClass
from simc_support.game_data.RaidRole import RaidRole
from simc_support.game_data.Role import Role
from simc_support.game_data.SimcObject import SimcObject
from simc_support.game_data.Stat import Stat
from simc_support.game_data.Talent import get_talent_dict


class WowSpec(SimcObject):
    """World of Warcraft Class spec data."""

    def __init__(
        self,
        id: int,
        wow_class: WowClass.WowClass,
        translations: Language.Translation,
        talents_blueprint: str,
        raid_role: RaidRole,
        role: Role,
        stat: Stat,
        ptr=False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.id = int(id)
        self.wow_class = wow_class

        if isinstance(translations, Language.Translation):
            self.translations = translations
        else:
            self.translations = Language.Translation(translations=translations)

        talents_blueprint = str(talents_blueprint)
        if len(talents_blueprint) != 7:
            raise ValueError("Wrong talent string length, expected 7")
        if (
            len(talents_blueprint.replace("1", "").replace("0", "").replace("x", ""))
            != 0
        ):
            raise ValueError(
                "Expected talent string to contain only 'x', '1' or '0' characters"
            )
        self.talents_blueprint = talents_blueprint

        if raid_role not in RaidRole:
            raise ValueError(f"Unknown raid_role '{raid_role}'")
        self.raid_role = raid_role

        if role not in Role:
            raise ValueError(f"Unknown role '{role}'")
        self.role = role

        if stat not in Stat:
            raise ValueError(f"Unknown stat '{stat}'")
        self.stat = stat

        self.talents = get_talent_dict(self, ptr)

    def is_dps_talent_combination(self, talent_combination: str) -> bool:
        """Determines whether a given talent combination is a valid talent combination.

        Arguments:
        talent_combination {str} -- [description]

        Returns:
        bool -- True, if all dps talent rows have a talent selected and all utility rows have no talent selected.
        """

        if not all([talent in "0123" for talent in talent_combination]):
            raise ValueError("Unexpected talent value. Values can be 0, 1, 2, 3.")

        for row, column in enumerate(self.talents_blueprint):
            if talent_combination[row] == "0" and (column == "1" or column == "x"):
                return False
            elif talent_combination[row] != "0" and column == "0":
                return False
        return True

    def get_talent_combinations(self, *, user_input: str = None) -> tuple:
        """Get a list of all valid dps talent combinations for a wow_class and wow_spec. If user_input is given, the provided mask is used for this genertion and pruning of the talent combination list.

        Arguments:

        Keyword Arguments:
            user_input {str} -- Mask to determine whether a combination is valid or not. Class inherent masks are used in addition to the user input. Input can either be empty, or len 7. len seven is understood as a mask for all talent rows. Write 'x' or '-' as a placeholder. Example: '201xx03' will generate 9 talent combinations if both x are dps talents. (default: {None})

        Returns:
            list[talent_combination{str}] -- List of all valied talent combinations for a class.
        """

        if not user_input:
            combinations = self._get_talent_combinations("xxxxxxx")
        elif len(user_input) == 7:
            combinations = self._get_talent_combinations(user_input)
        else:
            # something unexpected occured
            raise ValueError(f"Unexpected user_input '{user_input}'.")

        return tuple(combinations)

    def _get_talent_combinations(self, blueprint: str) -> tuple:
        """Generate all talent combinations matching blueprint. You're an enduser? Use get_talent_combinations(...).

        Arguments:
            blueprint {str} -- [description]

        Returns:
            list[talent_combination{str}] -- [description]
        """

        if not ("x" in blueprint or "-" in blueprint):
            return tuple(blueprint)

        blueprint = blueprint.replace("-", "x")

        pattern = ""

        for i, talent_type in enumerate(self.talents_blueprint):
            if blueprint[i] == "x" and talent_type == "0":
                pattern += "0"
            else:
                pattern += blueprint[i]

        raw_combinations = itertools.product(
            "0123",
            repeat=len(self.talents_blueprint),
        )

        combinations = map(
            lambda combination: "".join(combination),
            raw_combinations,
        )

        def filter_combination(combination: str) -> bool:
            """Return True if any combination position either doesn't match its pattern position or if any free position is not set to 1, 2, or 3.

            Args:
                combination (str): [description]

            Returns:
                bool: [description]
            """
            zipped = zip(pattern, combination)
            mapped = map(
                lambda t: (t[0] in "0123" and t[0] != t[1])
                or (t[0] == "x" and t[1] not in "123"),
                zipped,
            )
            return any(mapped)

        filtered_combinations = filter(
            lambda combination: not filter_combination(combination), combinations
        )

        return tuple(filtered_combinations)

    def __repr__(self) -> str:
        return " ".join([super().__repr__(), self.wow_class.__repr__()])

    def __str__(self) -> str:
        return " ".join([super().__str__(), self.wow_class.__str__()])

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "name": self.full_name,
            "wow_class": {
                "id": self.wow_class.id,
                "name": self.wow_class.full_name,
            },
        }


# Spec data here
BLOOD = WowSpec(
    id=250,
    wow_class=WowClass.DEATHKNIGHT,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1110011",
    raid_role=RaidRole.TANK,
    role=Role.MELEE,
    stat=Stat.STRENGTH,
    full_name="Blood",
    simc_name="blood",
)
FROST_DK = WowSpec(
    id=251,
    wow_class=WowClass.DEATHKNIGHT,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.STRENGTH,
    full_name="Frost",
    simc_name="frost",
)
UNHOLY = WowSpec(
    id=252,
    wow_class=WowClass.DEATHKNIGHT,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.STRENGTH,
    full_name="Unholy",
    simc_name="unholy",
)
HAVOC = WowSpec(
    id=577,
    wow_class=WowClass.DEMONHUNTER,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1110111",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Havoc",
    simc_name="havoc",
)
VENGEANCE = WowSpec(
    id=581,
    wow_class=WowClass.DEMONHUNTER,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1111111",
    raid_role=RaidRole.TANK,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Vengeance",
    simc_name="vengeance",
)
BALANCE = WowSpec(
    id=102,
    wow_class=WowClass.DRUID,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1000111",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Balance",
    simc_name="balance",
)
FERAL = WowSpec(
    id=103,
    wow_class=WowClass.DRUID,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1000111",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Feral",
    simc_name="feral",
)
GUARDIAN = WowSpec(
    id=104,
    wow_class=WowClass.DRUID,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1000111",
    raid_role=RaidRole.TANK,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Guardian",
    simc_name="guardian",
)
RESTORATION_DRUID = WowSpec(
    id=105,
    wow_class=WowClass.DRUID,
    translations=Language.EmptyTranslation(),
    talents_blueprint="0010000",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.INTELLECT,
    full_name="Restoration",
    simc_name="restoration",
)
BEASTMASTERY = WowSpec(
    id=253,
    wow_class=WowClass.HUNTER,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.AGILITY,
    full_name="Beast_Mastery",
    simc_name="beast_mastery",
)
MARKSMANSHIP = WowSpec(
    id=254,
    wow_class=WowClass.HUNTER,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.AGILITY,
    full_name="Marksmanship",
    simc_name="marksmanship",
)
SURVIVAL = WowSpec(
    id=255,
    wow_class=WowClass.HUNTER,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Survival",
    simc_name="survival",
)
ARCANE = WowSpec(
    id=62,
    wow_class=WowClass.MAGE,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1011011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Arcane",
    simc_name="arcane",
)
FIRE = WowSpec(
    id=63,
    wow_class=WowClass.MAGE,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1011011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Fire",
    simc_name="fire",
)
FROST_MAGE = WowSpec(
    id=64,
    wow_class=WowClass.MAGE,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1011011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Frost",
    simc_name="frost",
)
BREWMASTER = WowSpec(
    id=268,
    wow_class=WowClass.MONK,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010011",
    raid_role=RaidRole.TANK,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Brewmaster",
    simc_name="brewmaster",
)
WINDWALKER = WowSpec(
    id=269,
    wow_class=WowClass.MONK,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010011",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Windwalker",
    simc_name="windwalker",
)
MISTWEAVER = WowSpec(
    id=270,
    wow_class=WowClass.MONK,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1000011",
    raid_role=RaidRole.HEAL,
    role=Role.MELEE,
    stat=Stat.INTELLECT,
    full_name="Mistweaver",
    simc_name="mistweaver",
)
PROTECTION_PALADIN = WowSpec(
    id=66,
    wow_class=WowClass.PALADIN,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1100101",
    raid_role=RaidRole.TANK,
    role=Role.MELEE,
    stat=Stat.STRENGTH,
    full_name="Protection",
    simc_name="protection",
)
RETRIBUTION = WowSpec(
    id=70,
    wow_class=WowClass.PALADIN,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1100101",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.STRENGTH,
    full_name="Retribution",
    simc_name="retribution",
)
HOLY_PALADIN = WowSpec(
    id=65,
    wow_class=WowClass.PALADIN,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1100111",
    raid_role=RaidRole.HEAL,
    role=Role.MELEE,
    stat=Stat.INTELLECT,
    full_name="Holy",
    simc_name="holy",
)
DISCIPLINE = WowSpec(
    id=256,
    wow_class=WowClass.PRIEST,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010111",
    raid_role=RaidRole.HEAL,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Discipline",
    simc_name="discipline",
)
HOLY_PRIEST = WowSpec(
    id=257,
    wow_class=WowClass.PRIEST,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010111",
    raid_role=RaidRole.HEAL,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Holy",
    simc_name="holy",
)
SHADOW = WowSpec(
    id=258,
    wow_class=WowClass.PRIEST,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010111",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Shadow",
    simc_name="shadow",
)
ASSASSINATION = WowSpec(
    id=259,
    wow_class=WowClass.ROGUE,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1110011",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Assassination",
    simc_name="assassination",
)
OUTLAW = WowSpec(
    id=260,
    wow_class=WowClass.ROGUE,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010011",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Outlaw",
    simc_name="outlaw",
)
SUBTLETY = WowSpec(
    id=261,
    wow_class=WowClass.ROGUE,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1110011",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Subtlety",
    simc_name="subtlety",
)
ELEMENTAL = WowSpec(
    id=262,
    wow_class=WowClass.SHAMAN,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Elemental",
    simc_name="elemental",
)
ENHANCEMENT = WowSpec(
    id=263,
    wow_class=WowClass.SHAMAN,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.AGILITY,
    full_name="Enhancement",
    simc_name="enhancement",
)
RESTORATION_SHAMAN = WowSpec(
    id=264,
    wow_class=WowClass.SHAMAN,
    translations=Language.EmptyTranslation(),
    talents_blueprint="0100000",
    raid_role=RaidRole.HEAL,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Restoration",
    simc_name="restoration",
)
AFFLICTION = WowSpec(
    id=265,
    wow_class=WowClass.WARLOCK,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Affliction",
    simc_name="affliction",
)
DEMONOLOGY = WowSpec(
    id=266,
    wow_class=WowClass.WARLOCK,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Demonology",
    simc_name="demonology",
)
DESTRUCTION = WowSpec(
    id=267,
    wow_class=WowClass.WARLOCK,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1101011",
    raid_role=RaidRole.DD,
    role=Role.RANGED,
    stat=Stat.INTELLECT,
    full_name="Destruction",
    simc_name="destruction",
)
ARMS = WowSpec(
    id=71,
    wow_class=WowClass.WARRIOR,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010111",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.STRENGTH,
    full_name="Arms",
    simc_name="arms",
)
FURY = WowSpec(
    id=72,
    wow_class=WowClass.WARRIOR,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010111",
    raid_role=RaidRole.DD,
    role=Role.MELEE,
    stat=Stat.STRENGTH,
    full_name="Fury",
    simc_name="fury",
)
PROTECTION_WARRIOR = WowSpec(
    id=73,
    wow_class=WowClass.WARRIOR,
    translations=Language.EmptyTranslation(),
    talents_blueprint="1010111",
    raid_role=RaidRole.TANK,
    role=Role.MELEE,
    stat=Stat.STRENGTH,
    full_name="Protection",
    simc_name="protection",
)

WOWSPECS = [
    BLOOD,
    FROST_DK,
    UNHOLY,
    HAVOC,
    VENGEANCE,
    BALANCE,
    FERAL,
    GUARDIAN,
    RESTORATION_DRUID,
    BEASTMASTERY,
    MARKSMANSHIP,
    SURVIVAL,
    ARCANE,
    FIRE,
    FROST_MAGE,
    BREWMASTER,
    WINDWALKER,
    MISTWEAVER,
    PROTECTION_PALADIN,
    RETRIBUTION,
    HOLY_PALADIN,
    DISCIPLINE,
    HOLY_PRIEST,
    SHADOW,
    ASSASSINATION,
    OUTLAW,
    SUBTLETY,
    ELEMENTAL,
    ENHANCEMENT,
    RESTORATION_SHAMAN,
    AFFLICTION,
    DEMONOLOGY,
    DESTRUCTION,
    ARMS,
    FURY,
    PROTECTION_WARRIOR,
]


def get_wow_spec(
    wow_class: typing.Union[WowClass.WowClass, str], wow_spec: str
) -> WowSpec:
    """Obtain WowSpec from a WowClass/Name and wow spec name combination.

    Args:
        wow_class (typing.Union[WowClass.WowClass, str]): Instance of a WowClass or the name of that WowClass, e.g. "Mage"
        wow_spec (str): Name of the WowSpec, e.g. "Arcane"

    Raises:
        ValueError: An appropriate WowSpec couldn't be found.

    Returns:
        WowSpec
    """
    if isinstance(wow_class, str):
        wow_class = WowClass.get_wow_class(wow_class)

    for spec in WOWSPECS:
        if spec.wow_class == wow_class and wow_spec in (spec.full_name, spec.simc_name):
            return spec
    raise ValueError(
        f"No WowSpec found of class '{wow_class.simc_name}' and spec '{wow_spec}'."
    )


def get_wow_spec_from_id(wow_spec_id: int) -> WowSpec:
    """Obtain WoWSpec from a spec ID.

    Args:
        wow_spec_id (int): e.g. 62 (which would be Arcane Mage)

    Raises:
        ValueError: An appropriate WowSpec couldn't be found.

    Returns:
        WowSpec
    """
    for spec in WOWSPECS:
        if spec.id == wow_spec_id:
            return spec
    raise ValueError(f"No WowSpec found with id '{wow_spec_id}'.")


def get_wow_spec_from_combined_simc_name(wow_class_spec_name: str) -> WowSpec:
    """Obtain WoWSpec from a combined Class_Spec string from SimC with whitespace replaced by underscore.

    Args:
        wow_class_spec_name (str): e.g. "Death_Knight_Frost"

    Raises:
        ValueError: An appropriate WowSpec couldn't be found.

    Returns:
        WowSpec
    """
    for spec in WOWSPECS:
        combined_name = f"{spec.wow_class} {spec.full_name}".replace(" ", "_")
        if combined_name == wow_class_spec_name:
            return spec
    raise ValueError(f"No WowSpec found for class_spec '{wow_class_spec_name}'.")



import json
import pkg_resources
import typing

from simc_support.game_data.Covenant import Covenant
from simc_support.game_data.Covenant import COVENANTS
from simc_support.game_data.Language import _get_translations
from simc_support.game_data.Language import Translation
from simc_support.game_data.SimcObject import SimcObject
from simc_support.game_data.WowSpec import WowSpec
from simc_support.game_data.WowSpec import WOWSPECS


class Conduit(SimcObject):
    def __init__(
        self,
        *args,
        id: int,
        conduit_type: int,
        covenants: typing.List[Covenant],
        spec_mask: int,
        wow_specs: typing.List[WowSpec],
        spell_id: int,
        ranks: typing.List[int],
        translations: Translation,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.id = int(id)
        self.type = int(conduit_type)
        if not isinstance(covenants, (list, tuple)):
            raise TypeError(
                f"covenants expected list or tuple. Got '{type(covenants)}' instead."
            )
        if not all([isinstance(covenant, Covenant) for covenant in covenants]):
            raise TypeError(
                "covenants expected a list or tuple of Covenant type objects."
            )
        self.covenants = covenants
        self.spec_mask = int(spec_mask)
        if not all([isinstance(spec, WowSpec) for spec in wow_specs]):
            raise TypeError("wow_specs expected a list or tuple of WowSpec's.")
        self.wow_specs = wow_specs
        self.spell_id = int(spell_id)
        self.ranks = ranks

        if isinstance(translations, Translation):
            self.translations = translations
        elif isinstance(translations, dict):
            self.translations = Translation(translations=translations)
        else:
            raise TypeError(
                "translations must either be a dictionary or a Translaton object."
            )

    @property
    def is_finesse(self) -> bool:
        return self.type == 0

    @property
    def is_potency(self) -> bool:
        return self.type == 1

    @property
    def is_endurance(self) -> bool:
        return self.type == 2


def _load_conduits() -> typing.List[Conduit]:
    with pkg_resources.resource_stream(
        __name__, "/".join(("data_files", "conduits.json"))
    ) as f:
        loaded_conduits = json.load(f)

    def _get_specs(conduit: dict) -> typing.List[WowSpec]:
        return list([spec for spec in WOWSPECS if spec.id in conduit["spec_ids"]])

    def _get_covenants(conduit: dict) -> typing.List[Covenant]:
        return list(
            [
                covenant
                for covenant in COVENANTS
                if covenant.id == conduit["id_covenant"] or conduit["id_covenant"] == 0
            ]
        )

    conduits = []
    for conduit in loaded_conduits:
        conduits.append(
            Conduit(
                full_name=conduit["name"],
                id=conduit["id"],
                conduit_type=conduit["type"],
                covenants=_get_covenants(conduit),
                spec_mask=conduit["id_spec_set"],
                wow_specs=_get_specs(conduit),
                spell_id=conduit["spell_id"],
                ranks=list([rank + 1 for rank in conduit["ranks"]]),
                translations=_get_translations(conduit),
            )
        )
    return conduits


CONDUITS: typing.List[Conduit] = _load_conduits()


def get_conduits_for_spec(wow_spec: WowSpec) -> typing.Tuple[Conduit, ...]:
    """Return all dps conduits for a spec.

    Arguments:
      wow_spec {WowSpec} -- instance of a WowSpec

    Returns:
      tuple[Conduit] -- Tuple of all dps Conduits
    """

    conduits = [
        conduit
        for conduit in CONDUITS
        if wow_spec in conduit.wow_specs and conduit.is_potency
    ]

    return tuple(conduits)



import typing

class SimcObject(object):
    """Base class to enforce the usage of full_name for the User and
    simc_names in the background for SimulationCraft.
    """

    def __init__(self, full_name: str, simc_name: typing.Optional[str] = None, *args, **kwargs):
        super().__init__()
        self.full_name = full_name
        self._simc_name = simc_name

    @property
    def simc_name(self) -> str:
        if self._simc_name:
            return self._simc_name

        simc_name = self.full_name.lower()
        cleansers = (
            (" ", "_"),
            ("-", ""),
            ("'", ""),
            (":", ""),
            (",", ""),
        )

        for cleanser in cleansers:
            simc_name = simc_name.replace(*cleanser)

        self._simc_name = simc_name

        return self._simc_name

    def __str__(self) -> str:
        return self.full_name

    def __repr__(self) -> str:
        return self.simc_name



import enum


class Source(enum.Enum):
    CALLING = "Calling"
    DUNGEON = "Dungeon"
    MISSION = "Mission"
    PROFESSION = "Profession"
    PVP = "PvP"
    RAID = "Raid"
    RARE_MOB = "Rare Mob"
    UNKNOWN = "Unknown"
    WORLD_BOSS = "World Boss"
    WORLD_DROP = "World Drop"
    WORLD_QUEST = "World Quest"



BEASTLORD = "beastlord"
CASTINGPATCHWERK = "castingpatchwerk"
CLEAVEADD = "cleaveadd"
DUNGEONSLICE = "dungeonslice"
HEAVYMOVEMENT = "heavymovement"
HECTICADDCLEAVE = "hecticaddcleave"
HELTERSKELTER = "helterskelter"
LIGHTMOVEMENT = "lightmovement"
PATCHWERK = "patchwerk"
ULTRAXION = "ultraxion"


FIGHTSTYLES = (
    BEASTLORD,
    CASTINGPATCHWERK,
    CLEAVEADD,
    DUNGEONSLICE,
    HEAVYMOVEMENT,
    HECTICADDCLEAVE,
    HELTERSKELTER,
    LIGHTMOVEMENT,
    PATCHWERK,
    ULTRAXION,
)



import unittest

from simc_support.game_data.Legendary import Legendary
from simc_support.game_data.Legendary import LEGENDARIES


class TestLegendary(unittest.TestCase):
    def test_type(self):
        self.assertTrue(isinstance(LEGENDARIES, list))

    def test_non_emptyness(self):
        self.assertTrue(len(LEGENDARIES) > 0)

    def test_content_types(self):
        for legendary in LEGENDARIES:
            with self.subTest(legendary=legendary):
                self.assertTrue(isinstance(legendary, Legendary))



import typing

from simc_support.game_data.SimcObject import SimcObject
from simc_support.game_data import Race
from simc_support.game_data import Language


class WowClass(SimcObject):
    def __init__(
        self,
        id: int,
        races: typing.Tuple[Race.Race, ...],
        translations: Language.Translation,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.id = int(id)
        # double check input to be of Race class
        for race in races:
            if not isinstance(race, Race.Race):
                raise TypeError(
                    "Expected races to be of Race type. Got {} instead.".format(
                        type(race)
                    )
                )
            if race not in Race.RACES:
                raise ValueError("Unknown race {}.".format(race))
        self.races = races
        if isinstance(translations, Language.Translation):
            self.translations = translations
        else:
            self.translations = Language.Translation(translations=translations)


DEATHKNIGHT = WowClass(
    6,
    Race.RACES,
    Language.EmptyTranslation(),
    "Death Knight",
    "death_knight",
)
DEMONHUNTER = WowClass(
    12,
    (
        Race.NIGHTELF,
        Race.BLOODELF,
    ),
    Language.EmptyTranslation(),
    "Demon Hunter",
    "demon_hunter",
)
DRUID = WowClass(
    11,
    (
        Race.NIGHTELF,
        Race.WORGEN,
        Race.KULTIRAN,
        Race.TAUREN,
        Race.TROLL,
        Race.HIGHMOUNTAINTAUREN,
        Race.ZANDALARITROLL,
    ),
    Language.EmptyTranslation(),
    "Druid",
    "druid",
)
HUNTER = WowClass(3, Race.RACES, Language.EmptyTranslation(), "Hunter", "hunter")
MAGE = WowClass(
    8,
    (
        Race.DRAENEI,
        Race.DWARF,
        Race.GNOME,
        Race.HUMAN,
        Race.NIGHTELF,
        Race.PANDAREN_ALLIANCE,
        Race.WORGEN,
        Race.VOIDELF,
        Race.LIGHTFORGEDDRAENEI,
        Race.DARKIRONDWARF,
        Race.KULTIRAN,
        Race.MECHAGNOME,
        Race.BLOODELF,
        Race.GOBLIN,
        Race.ORC,
        Race.PANDAREN_HORDE,
        Race.TROLL,
        Race.UNDEAD,
        Race.NIGHTBORNE,
        Race.MAGHARORC,
        Race.ZANDALARITROLL,
        Race.VULPERA,
    ),
    Language.EmptyTranslation(),
    "Mage",
    "mage",
)
MONK = WowClass(
    10,
    (
        Race.DRAENEI,
        Race.DWARF,
        Race.GNOME,
        Race.HUMAN,
        Race.NIGHTELF,
        Race.PANDAREN_ALLIANCE,
        Race.VOIDELF,
        Race.DARKIRONDWARF,
        Race.KULTIRAN,
        Race.MECHAGNOME,
        Race.BLOODELF,
        Race.ORC,
        Race.PANDAREN_HORDE,
        Race.TAUREN,
        Race.TROLL,
        Race.UNDEAD,
        Race.NIGHTBORNE,
        Race.HIGHMOUNTAINTAUREN,
        Race.MAGHARORC,
        Race.ZANDALARITROLL,
        Race.VULPERA,
    ),
    Language.EmptyTranslation(),
    "Monk",
    "monk",
)
PALADIN = WowClass(
    2,
    (
        Race.DRAENEI,
        Race.DWARF,
        Race.HUMAN,
        Race.LIGHTFORGEDDRAENEI,
        Race.DARKIRONDWARF,
        Race.BLOODELF,
        Race.TAUREN,
        Race.ZANDALARITROLL,
    ),
    Language.EmptyTranslation(),
    "Paladin",
    "paladin",
)
PRIEST = WowClass(
    5,
    (
        Race.DRAENEI,
        Race.DWARF,
        Race.GNOME,
        Race.HUMAN,
        Race.NIGHTELF,
        Race.PANDAREN_ALLIANCE,
        Race.WORGEN,
        Race.VOIDELF,
        Race.LIGHTFORGEDDRAENEI,
        Race.DARKIRONDWARF,
        Race.KULTIRAN,
        Race.MECHAGNOME,
        Race.BLOODELF,
        Race.GOBLIN,
        Race.PANDAREN_HORDE,
        Race.TAUREN,
        Race.TROLL,
        Race.UNDEAD,
        Race.NIGHTBORNE,
        Race.MAGHARORC,
        Race.ZANDALARITROLL,
        Race.VULPERA,
    ),
    Language.EmptyTranslation(),
    "Priest",
    "priest",
)
ROGUE = WowClass(
    4,
    (
        Race.DWARF,
        Race.GNOME,
        Race.HUMAN,
        Race.NIGHTELF,
        Race.PANDAREN_ALLIANCE,
        Race.WORGEN,
        Race.VOIDELF,
        Race.DARKIRONDWARF,
        Race.KULTIRAN,
        Race.MECHAGNOME,
        Race.BLOODELF,
        Race.GOBLIN,
        Race.ORC,
        Race.PANDAREN_HORDE,
        Race.TROLL,
        Race.UNDEAD,
        Race.NIGHTBORNE,
        Race.MAGHARORC,
        Race.ZANDALARITROLL,
        Race.VULPERA,
    ),
    Language.EmptyTranslation(),
    "Rogue",
    "rogue",
)
SHAMAN = WowClass(
    7,
    (
        Race.DRAENEI,
        Race.DWARF,
        Race.PANDAREN_ALLIANCE,
        Race.DARKIRONDWARF,
        Race.KULTIRAN,
        Race.GOBLIN,
        Race.ORC,
        Race.PANDAREN_HORDE,
        Race.TAUREN,
        Race.TROLL,
        Race.HIGHMOUNTAINTAUREN,
        Race.MAGHARORC,
        Race.ZANDALARITROLL,
        Race.VULPERA,
    ),
    Language.EmptyTranslation(),
    "Shaman",
    "shaman",
)
WARLOCK = WowClass(
    9,
    (
        Race.DWARF,
        Race.GNOME,
        Race.HUMAN,
        Race.WORGEN,
        Race.VOIDELF,
        Race.DARKIRONDWARF,
        Race.MECHAGNOME,
        Race.BLOODELF,
        Race.GOBLIN,
        Race.ORC,
        Race.TROLL,
        Race.UNDEAD,
        Race.NIGHTBORNE,
        Race.VULPERA,
    ),
    Language.EmptyTranslation(),
    "Warlock",
    "warlock",
)
WARRIOR = WowClass(1, Race.RACES, Language.EmptyTranslation(), "Warrior", "warrior")

WOWCLASSES = (
    DEATHKNIGHT,
    DEMONHUNTER,
    DRUID,
    HUNTER,
    MAGE,
    MONK,
    PALADIN,
    PRIEST,
    ROGUE,
    SHAMAN,
    WARLOCK,
    WARRIOR,
)


def get_wow_class(name: str) -> WowClass:
    for wow_class in WOWCLASSES:
        if wow_class.full_name == name or wow_class.simc_name == name:
            return wow_class
    raise ValueError(f"No WowClass found with name '{name}'.")



import json
import pkg_resources
import typing

from simc_support.game_data.SimcObject import SimcObject
import simc_support.game_data.ItemLevel as ItemLevel
from simc_support.game_data.Language import _get_translations
from simc_support.game_data.Language import EmptyTranslation
from simc_support.game_data.Language import Translation
from simc_support.game_data.Role import Role
from simc_support.game_data.Source import Source
from simc_support.game_data.Stat import Stat
from simc_support.game_data.WowSpec import WowSpec


class Trinket(SimcObject):
    def __init__(
        self,
        *args,
        item_id: str,
        itemlevels: typing.List[int],
        role: typing.Optional[Role],
        stats: typing.Union[typing.List[Stat], typing.Tuple[Stat]],
        class_mask: int,
        translations: Translation,
        source: Source = None,
        on_use: bool = False,
        bonus_ids: typing.Union[typing.Tuple, typing.List[int], typing.Tuple[int]] = (),
        **kwargs,
    ):
        """Creates a Trinket instance

        Args:
            item_id (str): Item ID
            itemlevels (typing.List[int]): item is available at all these itemlevels
            role (Role):
            stats (typing.Union[typing.List[Stat], typing.Tuple[Stat]]): primary stats
            translations (Translation): name of the item in all languages
            source (str, optional): Drop source. Defaults to None.
            on_use (bool, optional): Is the trinket on use? Defaults to False.
            bonus_ids (typing.Union[typing.List[int], typing.Tuple[int]]):
        """
        super().__init__(translations.US, *args, **kwargs)
        self.translations = translations
        self.name: str = self.translations.US
        self.item_id: str = str(item_id)
        self.itemlevels: list = sorted(itemlevels)

        if isinstance(stats, list) or isinstance(stats, tuple):
            for element in stats:
                if element not in Stat:
                    raise TypeError("One or more provided stats are unknown.")
        else:
            raise TypeError(f"stats: Expected list or tuple. Got {type(stats)}")
        if isinstance(stats, list) or isinstance(stats, tuple):
            self.stats = tuple(stats)

        self.class_mask = class_mask
        self.role = role

        if not isinstance(source, Source):
            raise ValueError(f"Unknown source '{source}'.")
        self.source: Source = source

        self.on_use: bool = bool(on_use)

        if isinstance(bonus_ids, list) or isinstance(bonus_ids, tuple):
            for bonus in bonus_ids:
                if not isinstance(bonus, int):
                    raise TypeError("One or more provided bonus IDs was not an INT.")
        else:
            raise TypeError(f"bonus_id: Expected list or tuple. Got {type(bonus_ids)}")
        if isinstance(bonus_ids, list) or isinstance(bonus_ids, tuple):
            self.bonus_ids = tuple(bonus_ids)

    @property
    def min_itemlevel(self) -> int:
        return self.itemlevels[0]

    @property
    def max_itemlevel(self) -> int:
        return self.itemlevels[-1]

    def is_usable_by_class(self, class_id: int) -> bool:
        return bool(self.class_mask & 1 << (class_id - 1))

    def _stringify(self) -> str:
        return f"{self.item_id} {self.name} {self.stats} {self.source}"


def _load_trinkets() -> typing.List[Trinket]:
    with pkg_resources.resource_stream(
        __name__, "/".join(("data_files", "trinkets.json"))
    ) as f:
        loaded_trinkets = json.load(f)

    def _get_stats(item: dict) -> typing.List[Stat]:
        """Get primary stats from items stat_type information.
        TODO: Can be extended using ItemEffect.id_specialization.
        TODO: what to do with trinkets without primary stats?

        Args:
            item (dict): [description]

        Returns:
            typing.List[Stat]: [description]
        """
        stats = []
        translation = {3: Stat.AGILITY, 4: Stat.STRENGTH, 5: Stat.INTELLECT}
        for i in range(1, 11):
            if item[f"stat_type_{i}"] in translation:
                stats.append(translation[item[f"stat_type_{i}"]])
            elif item[f"stat_type_{i}"] == 71:
                stats.append(Stat.AGILITY)
                stats.append(Stat.STRENGTH)
                stats.append(Stat.INTELLECT)
            elif item[f"stat_type_{i}"] == 72:
                stats.append(Stat.AGILITY)
                stats.append(Stat.STRENGTH)
            elif item[f"stat_type_{i}"] == 73:
                stats.append(Stat.AGILITY)
                stats.append(Stat.INTELLECT)
            elif item[f"stat_type_{i}"] == 74:
                stats.append(Stat.STRENGTH)
                stats.append(Stat.INTELLECT)
        return stats

    def _get_itemlevels(item: dict, source: Source) -> typing.List[int]:

        # add special cases here
        if item["name"] == "Spiritual Alchemy Stone":
            return [
                item.get("ilevel", 0),
                200,
            ]

        # Tazavesh Dungeon m0 specific itemlevel
        if item["id_journal_instance"] == 1194:
            return [ItemLevel.TAZAVESH]

        if source == Source.DUNGEON:
            return ItemLevel.DUNGEON
        if source == Source.RAID and item["id_journal_instance"] == 1190:
            if item["id"] in (
                184027,  # Stone Legion Heraldry
                184028,  # Cabalist's Hymnal
                184030,  # Dreadfire Vessel
                184029,  # Manabound Mirror
                184031,  # Sanguine Vintage
            ):
                return ItemLevel.CASTLE_NATHRIA_ENDBOSSES
            else:
                return ItemLevel.CASTLE_NATHRIA
        if source == Source.PROFESSION:
            return [
                item["ilevel"],
            ]
        if source == Source.PVP:
            if "Aspirant" in item["name"]:
                return [
                    itemlevel
                    for itemlevel in ItemLevel.PVP_HONOR
                    if itemlevel >= item.get("ilevel", 0)
                ]
            else:
                return [
                    itemlevel
                    for itemlevel in ItemLevel.PVP_CONQUEST
                    if itemlevel >= item.get("ilevel", 0)
                ]
        if source == Source.RARE_MOB:
            return ItemLevel.RARE_MOB.get(item["ilevel"], [item["ilevel"]])

        if source == Source.CALLING:
            return ItemLevel.CALLINGS

        return [
            item["ilevel"],
        ]

    def _get_bonus_ids(item: dict) -> typing.List[int]:
        ids = []

        # Unbound Changeling
        if item["id"] == 178708:
            # crit
            ids.append(6916)
            # haste
            ids.append(6917)
            # mastery
            ids.append(6918)
            # crit haste mastery
            ids.append(6915)

        return ids

    def _get_source(item: dict) -> Source:
        instance_mapping = {
            1: Source.DUNGEON,
            2: Source.RAID,
        }
        item_mapping = {
            175884: Source.PVP,
            175921: Source.PVP,
            178298: Source.PVP,
            178334: Source.PVP,
            178386: Source.PVP,
            178447: Source.PVP,
            181333: Source.PVP,
            181335: Source.PVP,
            181816: Source.PVP,
            184052: Source.PVP,
            184053: Source.PVP,
            184054: Source.PVP,
            184055: Source.PVP,
            184056: Source.PVP,
            184057: Source.PVP,
            184058: Source.PVP,
            184059: Source.PVP,
            184060: Source.PVP,
            184807: Source.WORLD_DROP,
            182455: Source.RARE_MOB,
            182454: Source.RARE_MOB,
            182453: Source.RARE_MOB,
            182452: Source.RARE_MOB,
            182451: Source.RARE_MOB,
            175729: Source.RARE_MOB,
            175943: Source.PROFESSION,
            175942: Source.PROFESSION,
            175941: Source.PROFESSION,
            171323: Source.PROFESSION,
            173069: Source.PROFESSION,
            173078: Source.PROFESSION,
            173087: Source.PROFESSION,
            173096: Source.PROFESSION,
            175728: Source.CALLING,
            181334: Source.CALLING,
            181357: Source.CALLING,
            181358: Source.CALLING,
            181359: Source.CALLING,
            181360: Source.CALLING,
            181457: Source.CALLING,
            181458: Source.CALLING,
            181459: Source.CALLING,
            181501: Source.CALLING,
            181502: Source.CALLING,
            181503: Source.CALLING,
            181507: Source.CALLING,
            184839: Source.CALLING,
            184840: Source.CALLING,
            184841: Source.CALLING,
            184842: Source.CALLING,
        }
        if item["instance_type"]:
            # use handcrafted mapping
            return instance_mapping[item["instance_type"]]
        # print(item)
        return item_mapping.get(item["id"], Source.UNKNOWN)

    trinkets = []
    for trinket in loaded_trinkets:
        source = _get_source(trinket)
        itemlevels = _get_itemlevels(trinket, source)
        if not itemlevels:
            continue
        trinkets.append(
            Trinket(
                item_id=trinket["id"],
                itemlevels=itemlevels,
                role=None,  # TODO
                stats=_get_stats(trinket),
                class_mask=trinket["class_mask"],
                translations=_get_translations(trinket),
                source=_get_source(trinket),
                on_use=trinket["on_use"],
                bonus_ids=_get_bonus_ids(trinket),
            )
        )
    return trinkets


TRINKETS: typing.List[Trinket] = _load_trinkets()


def get_trinkets_for_spec(wow_spec: WowSpec) -> typing.Tuple[Trinket, ...]:
    """New function to return all available trinkets for a spec

    Arguments:
      wow_spec {WowSpec} -- instance of a WowSpec

    Returns:
      tuple[Trinket] -- Tuple of all Trinkets
    """

    trinkets: typing.List[Trinket] = []
    for trinket in TRINKETS:
        if trinket.is_usable_by_class(wow_spec.wow_class.id):
            if wow_spec.stat in trinket.stats or len(trinket.stats) == 0:
                trinkets.append(trinket)

    return tuple(trinkets)


def get_versatility_trinket(stat: Stat) -> Trinket:
    """Returns a vers stat stick with the given Stat.

    Returns:
        Trinket -- Versatility Stat stick
    """
    empty_translation = EmptyTranslation()
    empty_translation.US = "Versatility Stat Stick"
    all_classes = 2 ** 16 - 1
    if stat == Stat.AGILITY:
        # "Stat Stick (Versatility)", "142506,bonus_id=607"
        return Trinket(
            item_id="142506",
            bonus_ids=[
                607,
            ],
            itemlevels=[ItemLevel.CASTLE_NATHRIA[0]],
            role=None,
            stats=[
                Stat.AGILITY,
            ],
            class_mask=all_classes,
            translations=empty_translation,
            source=Source.UNKNOWN,
            on_use=False,
        )
    elif stat == Stat.INTELLECT:
        # "Stat Stick (Versatility)", "142507,bonus_id=607"
        return Trinket(
            item_id="142507",
            bonus_ids=[
                607,
            ],
            itemlevels=[ItemLevel.CASTLE_NATHRIA[0]],
            role=None,
            stats=[
                Stat.INTELLECT,
            ],
            class_mask=all_classes,
            translations=empty_translation,
            source=Source.UNKNOWN,
            on_use=False,
        )
    elif stat == Stat.STRENGTH:
        # "Stat Stick (Versatility)", "142508,bonus_id=607"
        return Trinket(
            item_id="142508",
            bonus_ids=[
                607,
            ],
            itemlevels=[ItemLevel.CASTLE_NATHRIA[0]],
            role=None,
            stats=[
                Stat.STRENGTH,
            ],
            class_mask=all_classes,
            translations=empty_translation,
            source=Source.UNKNOWN,
            on_use=False,
        )
    raise ValueError(f"Unknown stat {stat}. No Versatility trinket available.")


def _get_trinket(id: int = None, name: str = None) -> Trinket:
    if not (id or name):
        raise ValueError("id or name must be provided to look up a trinket")

    for trinket in TRINKETS:
        if trinket.item_id == id or trinket.name == name:
            return trinket
    raise FileNotFoundError("Trinket not found")


def get_trinket(search: typing.Union[int, str]) -> Trinket:
    if isinstance(search, int):
        return _get_trinket(id=search)
    elif isinstance(search, str):
        return _get_trinket(name=search)



import typing

from simc_support.game_data import Faction
from simc_support.game_data.Faction import ALLIANCE
from simc_support.game_data.Faction import HORDE
from simc_support.game_data.Language import Translation, Language
from simc_support.game_data.SimcObject import SimcObject


class Race(SimcObject):
    """Race defines how your character looks, moves, and what racial ability you have."""

    def __init__(
        self,
        faction: Faction.Faction,
        full_name: str,
        simc_name: str,
        translations: typing.Union[typing.Dict, Translation],
        *args,
        **kwargs,
    ) -> None:
        super().__init__(full_name, simc_name, *args, **kwargs)

        self.faction = faction

        if isinstance(translations, Translation):
            self.translations = translations
        elif isinstance(translations, dict):
            self.translations = Translation(translations=translations)
        else:
            raise TypeError(
                "translations must either be a dictionary or a Translaton object."
            )


# Alliance
DARKIRONDWARF = Race(
    ALLIANCE,
    "Dark Iron Dwarf",
    "dark_iron_dwarf",
    {
        Language.US: "Dark Iron Dwarf",
        Language.IT: "Nano Ferroscuro",
        Language.DE: "Dunkeleisenzwerg",
        Language.FR: "Nain sombrefer",
        Language.RU: "Дворф из клана Черного Железа",
        Language.ES: "Enano Hierro Negro",
        Language.KR: "검은무쇠 드워프",
        Language.CN: "黑铁矮人",
        Language.BR: "Anão Ferro Negro",
    },
)
DRAENEI = Race(
    ALLIANCE,
    "Draenei",
    "draenei",
    {
        Language.US: "Draenei",
        Language.IT: "Draenei",
        Language.DE: "Draenei",
        Language.FR: "Draeneï",
        Language.RU: "Дреней",
        Language.ES: "Draenei",
        Language.KR: "드레나이",
        Language.CN: "德莱尼",
        Language.BR: "Draenei",
    },
)
DWARF = Race(
    ALLIANCE,
    "Dwarf",
    "dwarf",
    {
        Language.US: "Dwarf",
        Language.IT: "Nano",
        Language.DE: "Zwerg",
        Language.FR: "Nain",
        Language.RU: "Дворф",
        Language.ES: "Enano",
        Language.KR: "드워프",
        Language.CN: "矮人",
        Language.BR: "Anão",
    },
)
GNOME = Race(
    ALLIANCE,
    "Gnome",
    "gnome",
    {
        Language.US: "Gnome",
        Language.IT: "Gnomo",
        Language.DE: "Gnom",
        Language.FR: "Gnome",
        Language.RU: "Гном",
        Language.ES: "Gnomo",
        Language.KR: "노움",
        Language.CN: "侏儒",
        Language.BR: "Gnomo",
    },
)
HUMAN = Race(
    ALLIANCE,
    "Human",
    "human",
    {
        Language.US: "Human",
        Language.IT: "Umano",
        Language.DE: "Mensch",
        Language.FR: "Humain",
        Language.RU: "Человек",
        Language.ES: "Humano",
        Language.KR: "인간",
        Language.CN: "人类",
        Language.BR: "Humano",
    },
)
KULTIRAN = Race(
    ALLIANCE,
    "Kul Tiran",
    "kul_tiran",
    {
        Language.US: "Kul Tiran",
        Language.IT: "Kul Tirano",
        Language.DE: "Kul Tiraner",
        Language.FR: "Kultirassien",
        Language.RU: "Култирасец",
        Language.ES: "Ciudadano de Kul Tiras",
        Language.KR: "쿨 티란",
        Language.CN: "库尔提拉斯人",
        Language.BR: "Kultireno",
    },
)
LIGHTFORGEDDRAENEI = Race(
    ALLIANCE,
    "Lightforged Draenei",
    "lightforged_draenei",
    {
        Language.US: "Lightforged Draenei",
        Language.IT: "Draenei Forgialuce",
        Language.DE: "Lichtgeschmiedeter Draenei",
        Language.FR: "Draeneï sancteforge",
        Language.RU: "Озаренный дреней",
        Language.ES: "Draenei forjado por la Luz",
        Language.KR: "빛벼림 드레나이",
        Language.CN: "光铸德莱尼",
        Language.BR: "Draenei Forjado a Luz",
    },
)
MECHAGNOME = Race(
    ALLIANCE,
    "Mechagnome",
    "mechagnome",
    {
        Language.US: "Mechagnome",
        Language.IT: "Meccagnomo",
        Language.DE: "Mechagnom",
        Language.FR: "Mécagnome",
        Language.RU: "Механогном",
        Language.ES: "Mecagnomo",
        Language.KR: "기계노움",
        Language.CN: "机械侏儒",
        Language.BR: "Gnomecânico",
    },
)
NIGHTELF = Race(
    ALLIANCE,
    "Night Elf",
    "night_elf",
    {
        Language.US: "Night Elf",
        Language.IT: "Elfo della Notte",
        Language.DE: "Nachtelf",
        Language.FR: "Elfe de la nuit",
        Language.RU: "Ночной эльф",
        Language.ES: "Elfo de la noche",
        Language.KR: "나이트 엘프",
        Language.CN: "暗夜精灵",
        Language.BR: "Elfo Noturno",
    },
)
PANDAREN_ALLIANCE = Race(
    ALLIANCE,
    "Pandaren",
    "pandaren",
    {
        Language.US: "Pandaren",
        Language.IT: "Pandaren",
        Language.DE: "Pandaren",
        Language.FR: "Pandaren",
        Language.RU: "Пандарен",
        Language.ES: "Pandaren",
        Language.KR: "판다렌",
        Language.CN: "熊猫人",
        Language.BR: "Pandaren",
    },
)
WORGEN = Race(
    ALLIANCE,
    "Worgen",
    "worgen",
    {
        Language.US: "Worgen",
        Language.IT: "Worgen",
        Language.DE: "Worgen",
        Language.FR: "Worgen",
        Language.RU: "Ворген",
        Language.ES: "Huargen",
        Language.KR: "늑대인간",
        Language.CN: "狼人",
        Language.BR: "Worgen",
    },
)
VOIDELF = Race(
    ALLIANCE,
    "Void Elf",
    "void_elf",
    {
        Language.US: "Void Elf",
        Language.IT: "Elfo del Vuoto",
        Language.DE: "Leerenelf",
        Language.FR: "Elfe du Vide",
        Language.RU: "Эльф Бездны",
        Language.ES: "Elfo del Vacío",
        Language.KR: "공허 엘프",
        Language.CN: "虚空精灵",
        Language.BR: "Elfo Caótico",
    },
)

# Horde
BLOODELF = Race(
    HORDE,
    "Blood Elf",
    "blood_elf",
    {
        Language.US: "Blood Elf",
        Language.IT: "Elfo del Sangue",
        Language.DE: "Blutelf",
        Language.FR: "Elfe de sang",
        Language.RU: "Эльф крови",
        Language.ES: "Elfo de sangre",
        Language.KR: "블러드 엘프",
        Language.CN: "血精灵",
        Language.BR: "Elfo Sangrento",
    },
)
GOBLIN = Race(
    HORDE,
    "Goblin",
    "goblin",
    {
        Language.US: "Goblin",
        Language.IT: "Goblin",
        Language.DE: "Goblin",
        Language.FR: "Gobelin",
        Language.RU: "Гоблин",
        Language.ES: "Goblin",
        Language.KR: "고블린",
        Language.CN: "地精",
        Language.BR: "Goblin",
    },
)
HIGHMOUNTAINTAUREN = Race(
    HORDE,
    "Hightmountain Tauren",
    "highmountain_tauren",
    {
        Language.US: "Highmountain Tauren",
        Language.IT: "Tauren di Alto Monte",
        Language.DE: "Hochbergtauren",
        Language.FR: "Tauren de Haut-Roc",
        Language.RU: "Таурен Крутогорья",
        Language.ES: "Tauren Monte Alto",
        Language.KR: "높은산 타우렌",
        Language.CN: "至高岭牛头人",
        Language.BR: "Tauren Altamontês",
    },
)
MAGHARORC = Race(
    HORDE,
    "Mag'har Orc",
    "maghar_orc",
    {
        Language.US: "Mag'har Orc",
        Language.IT: "Orco Mag'har",
        Language.DE: "Mag'har",
        Language.FR: "Orc mag’har",
        Language.RU: "Маг'хар",
        Language.ES: "Orco Mag'har",
        Language.KR: "마그하르 오크",
        Language.CN: "玛格汉兽人",
        Language.BR: "Orc Mag'har",
    },
)
NIGHTBORNE = Race(
    HORDE,
    "Nightborne",
    "nightborne",
    {
        Language.US: "Nightborne",
        Language.IT: "Nobile Oscuro",
        Language.DE: "Nachtgeborener",
        Language.FR: "Sacrenuit",
        Language.RU: "Ночнорожденный",
        Language.ES: "Nocheterna",
        Language.KR: "나이트본",
        Language.CN: "夜之子",
        Language.BR: "Filho da Noite",
    },
)
ORC = Race(
    HORDE,
    "Orc",
    "orc",
    {
        Language.US: "Orc",
        Language.IT: "Orco",
        Language.DE: "Orc",
        Language.FR: "Orc",
        Language.RU: "Орк",
        Language.ES: "Orco",
        Language.KR: "오크",
        Language.CN: "兽人",
        Language.BR: "Orc",
    },
)
PANDAREN_HORDE = Race(
    HORDE,
    "Pandaren",
    "pandaren",
    {
        Language.US: "Pandaren",
        Language.IT: "Pandaren",
        Language.DE: "Pandaren",
        Language.FR: "Pandaren",
        Language.RU: "Пандарен",
        Language.ES: "Pandaren",
        Language.KR: "판다렌",
        Language.CN: "熊猫人",
        Language.BR: "Pandaren",
    },
)
TAUREN = Race(
    HORDE,
    "Tauren",
    "tauren",
    {
        Language.US: "Tauren",
        Language.IT: "Tauren",
        Language.DE: "Tauren",
        Language.FR: "Tauren",
        Language.RU: "Таурен",
        Language.ES: "Tauren",
        Language.KR: "타우렌",
        Language.CN: "牛头人",
        Language.BR: "Tauren",
    },
)
TROLL = Race(
    HORDE,
    "Troll",
    "troll",
    {
        Language.US: "Troll",
        Language.IT: "Troll",
        Language.DE: "Troll",
        Language.FR: "Troll",
        Language.RU: "Тролль",
        Language.ES: "Trol",
        Language.KR: "트롤",
        Language.CN: "巨魔",
        Language.BR: "Troll",
    },
)
UNDEAD = Race(
    HORDE,
    "Undead",
    "undead",
    {
        Language.US: "Undead",
        Language.IT: "Non Morto",
        Language.DE: "Untoter",
        Language.FR: "Mort-vivant",
        Language.RU: "Нежить",
        Language.ES: "No-muerto",
        Language.KR: "언데드",
        Language.CN: "亡灵",
        Language.BR: "Morto-vivo",
    },
)
VULPERA = Race(
    HORDE,
    "Vulpera",
    "vulpera",
    {
        Language.US: "Vulpera",
        Language.IT: "Vulpera",
        Language.DE: "Vulpera",
        Language.FR: "Vulpérin",
        Language.RU: "Вульпера",
        Language.ES: "Vulpera",
        Language.KR: "불페라",
        Language.CN: "狐人",
        Language.BR: "Vulpera",
    },
)
ZANDALARITROLL = Race(
    HORDE,
    "Zandalari Troll",
    "zandalari_troll",
    {
        Language.US: "Zandalari Troll",
        Language.IT: "Troll Zandalari",
        Language.DE: "Zandalaritroll",
        Language.FR: "Troll zandalari",
        Language.RU: "Зандалар",
        Language.ES: "Trol Zandalari",
        Language.KR: "잔달라 트롤",
        Language.CN: "赞达拉巨魔",
        Language.BR: "Troll Zandalari",
    },
)

RACES = (
    DARKIRONDWARF,
    DRAENEI,
    DWARF,
    GNOME,
    HUMAN,
    KULTIRAN,
    LIGHTFORGEDDRAENEI,
    MECHAGNOME,
    NIGHTELF,
    PANDAREN_ALLIANCE,
    WORGEN,
    VOIDELF,
    BLOODELF,
    GOBLIN,
    HIGHMOUNTAINTAUREN,
    MAGHARORC,
    NIGHTBORNE,
    ORC,
    PANDAREN_HORDE,
    TAUREN,
    TROLL,
    UNDEAD,
    VULPERA,
    ZANDALARITROLL,
)


def get_race(name: str) -> Race:
    for race in RACES:
        if race.full_name == name or race.simc_name == name:
            return race
    raise ValueError(f"No race '{name}' found.")



import json
import pkg_resources
import typing

from simc_support.game_data.SimcObject import SimcObject
from simc_support.game_data.Language import Translation, _get_translations
from simc_support.game_data.WowSpec import WowSpec, WOWSPECS

# Legendary IDs which can be used by different specs than game data suggests.
_EXCEPTIONS = {
    9: [63, 64],  # Temporal Warp
    10: [62, 64],  # Fevered Incantation
    42: [250, 251],  # Reanimated Shambler
    44: [250],  # Death's Certainty
    66: [112],  # Vanguard's Momentum
    70: [255],  # Dire Command
    89: [269],  # Shaohao's Might
    95: [268],  # Keefer's Skyreach
    111: [66],  # Relentless Inquisitor
    113: [259],  # Finality
    129: [259, 260],  # Deathly Shadows
    135: [263],  # Elemental Equilibrium
    # 152: [258], # Kiss of Death # Shadow, needs checking
}


class Legendary(SimcObject):
    """World of Warcraft Legendary class"""

    def __init__(
        self,
        *args,
        id: int,
        bonus_id: int,
        spell_id: int,
        wow_specs: typing.List[WowSpec],
        translations: Translation,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.id = int(id)
        self.bonus_id = int(bonus_id)
        self.spell_id = int(spell_id)

        if not all([isinstance(spec, WowSpec) for spec in wow_specs]):
            raise TypeError("wow_specs expected a list or tuple of WowSpec's.")
        self.wow_specs = wow_specs

        if isinstance(translations, Translation):
            self.translations = translations
        else:
            self.translations = Translation(translations=translations)


def _load_legendaries() -> typing.List[Legendary]:
    with pkg_resources.resource_stream(
        __name__, "/".join(("data_files", "legendaries.json"))
    ) as f:
        loaded_legendaries = json.load(f)

    def _get_specs(legendary) -> typing.List[WowSpec]:
        specs: typing.List[int] = legendary["spec_ids"]

        wow_specs: typing.List[WowSpec] = []

        for spec in specs:
            for real_spec in WOWSPECS:
                if spec == real_spec.id:
                    wow_specs.append(real_spec)

        return wow_specs

    legendaries = []
    for legendary in loaded_legendaries:
        legendaries.append(
            Legendary(
                full_name=legendary["name"],
                id=legendary["id"],
                spell_id=legendary["id_spell"],
                bonus_id=legendary["id_bonus"],
                translations=_get_translations(legendary),
                wow_specs=_get_specs(legendary),
            )
        )

    return legendaries


LEGENDARIES: typing.List[Legendary] = _load_legendaries()


def get_legendaries_for_spec(wow_spec: WowSpec) -> typing.Tuple[Legendary, ...]:
    """Return a tuple of legendaries for the provided WowSpec.

    Args:
        wow_spec (WowSpec): a valid WowSpec class instance

    Returns:
        typing.List[Legendary]: [description]
    """
    if not isinstance(wow_spec, WowSpec):
        raise TypeError("wow_spec needs to be of type WowSpec.")
    return tuple(
        [
            legendary
            for legendary in LEGENDARIES
            if wow_spec in legendary.wow_specs
            or wow_spec.id in _EXCEPTIONS.get(legendary.id, [])
        ]
    )



import copy
import json
import pkg_resources
import typing

from simc_support.game_data.Language import Translation, _get_translations
from simc_support.game_data.SimcObject import SimcObject
from simc_support.game_data.Covenant import Covenant, get_covenant


SECOND_SOULBIND_RENOWN = 6
THIRD_SOULBIND_RENOWN = 25
# until found/extracted in game data, here the rown table
# soulbind_id: tier: renown
RENOWN = {
    1: {  # Niya
        0: 0,
        1: 0,
        2: 3,
        3: 7,
        4: 10,
        5: 18,
        6: 25,
        7: 30,
    },
    2: {  # Dreamweaver
        0: SECOND_SOULBIND_RENOWN,
        1: SECOND_SOULBIND_RENOWN,
        2: SECOND_SOULBIND_RENOWN,
        3: 9,
        4: 13,
        5: 21,
        6: 29,
        7: 34,
    },
    3: {  # General Draven
        0: THIRD_SOULBIND_RENOWN,
        1: THIRD_SOULBIND_RENOWN,
        2: THIRD_SOULBIND_RENOWN,
        3: THIRD_SOULBIND_RENOWN,
        4: THIRD_SOULBIND_RENOWN,
        5: THIRD_SOULBIND_RENOWN,
        6: 29,
        7: 31,
    },
    4: {  # Plague Deviser Marileth
        0: 0,
        1: 0,
        2: 3,
        3: 9,
        4: 10,
        5: 21,
        6: 25,
        7: 30,
    },
    5: {  # Emeni
        0: SECOND_SOULBIND_RENOWN,
        1: SECOND_SOULBIND_RENOWN,
        2: SECOND_SOULBIND_RENOWN,
        3: 7,
        4: 13,
        5: 18,
        6: 28,
        7: 34,
    },
    6: {  # Korayn
        0: THIRD_SOULBIND_RENOWN,
        1: THIRD_SOULBIND_RENOWN,
        2: THIRD_SOULBIND_RENOWN,
        3: THIRD_SOULBIND_RENOWN,
        4: THIRD_SOULBIND_RENOWN,
        5: THIRD_SOULBIND_RENOWN,
        6: 28,
        7: 31,
    },
    7: {  # Pelagos
        0: 0,
        1: 0,
        2: 3,
        3: 7,
        4: 10,
        5: 21,
        6: 25,
        7: 30,
    },
    8: {  # Nadjia the Mistblade
        0: 0,
        1: 0,
        2: 3,
        3: 7,
        4: 10,
        5: 21,
        6: 25,
        7: 30,
    },
    9: {  # Theotar the Mad Duke
        0: SECOND_SOULBIND_RENOWN,
        1: SECOND_SOULBIND_RENOWN,
        2: SECOND_SOULBIND_RENOWN,
        3: 9,
        4: 13,
        5: 18,
        6: 28,
        7: 34,
    },
    10: {  # Bonesmith Heirmir
        0: THIRD_SOULBIND_RENOWN,
        1: THIRD_SOULBIND_RENOWN,
        2: THIRD_SOULBIND_RENOWN,
        3: THIRD_SOULBIND_RENOWN,
        4: THIRD_SOULBIND_RENOWN,
        5: THIRD_SOULBIND_RENOWN,
        6: 29,
        7: 31,
    },
    13: {  # Kleia
        0: SECOND_SOULBIND_RENOWN,
        1: SECOND_SOULBIND_RENOWN,
        2: SECOND_SOULBIND_RENOWN,
        3: 9,
        4: 13,
        5: 18,
        6: 28,
        7: 34,
    },
    18: {  # Forgelite Prime Mikanikos
        0: THIRD_SOULBIND_RENOWN,
        1: THIRD_SOULBIND_RENOWN,
        2: THIRD_SOULBIND_RENOWN,
        3: THIRD_SOULBIND_RENOWN,
        4: THIRD_SOULBIND_RENOWN,
        5: THIRD_SOULBIND_RENOWN,
        6: 29,
        7: 31,
    },
}


class SoulBindTalent(SimcObject):
    def __init__(
        self,
        *args,
        id: int,
        spell_id: int,
        tier: int,
        order: int,
        parent_id: int,
        soulbind_id: int,
        type: int,
        translations: typing.Union[typing.Dict, Translation],
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.id = id
        self.spell_id = spell_id
        self.tier = tier
        self.order = order
        self.parent_id = parent_id
        self.soulbind_id = soulbind_id
        self.type = type
        if isinstance(translations, Translation):
            self.translations = translations
        elif isinstance(translations, dict):
            self.translations = Translation(translations=translations)
        else:
            raise TypeError(
                "translations must either be a dictionary or a Translaton object."
            )

    @property
    def is_dps_increase(self) -> bool:
        WHITELIST = [
            # Pelagos
            "Combat Meditation",
            # Kleia
            "Pointed Courage",
            # Forgelite Prime Mikanikos
            "Hammer of Genesis",
            "Bron's Call to Action",
            # Plague Deviser Marelith
            "Plaguey's Preemptive Strike",
            # Emeni
            "Lead by Example",
            "Gnashing Chompers",
            # Bonesmith Heirmir
            "Forgeborne Reveries",
            "Heirmir's Arsenal: Marrowed Gemstone",
            # Niya
            "Niya's Tools: Burrs",
            "Niya's Tools: Poison",
            "Grove Invigoration",
            # Dreamweaver
            "Field of Blossoms",
            "Social Butterfly",
            # Korayn
            "First Strike",
            "Wild Hunt Tactics",
            # Nadja the Mistblade
            "Exacting Preparation",
            "Dauntless Duelist",
            "Thrill Seeker",
            # Theotar the Mad Duke
            "Refined Palate",
            "Soothing Shade",
            "Wasteland Propriety",
            # General Draven
            "Superior Tactics",
            "Built for War",
        ]
        return self.full_name in WHITELIST or self.is_potency

    @property
    def is_finesse(self) -> bool:
        return self.type == 1

    @property
    def is_potency(self) -> bool:
        return self.type == 2

    @property
    def is_endurance(self) -> bool:
        return self.type == 3

    @property
    def renown(self) -> int:
        return RENOWN[self.soulbind_id][self.tier]


class SoulBind(SimcObject):
    """Shadowlands specific additional talent tree."""

    def __init__(
        self,
        *args,
        id: int,
        covenant: Covenant,
        translations: typing.Union[typing.Dict, Translation],
        soul_bind_talents: typing.List[SoulBindTalent],
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.id = id
        self.covenant = covenant
        self.soul_bind_talents = soul_bind_talents

        if isinstance(translations, Translation):
            self.translations = translations
        elif isinstance(translations, dict):
            self.translations = Translation(translations=translations)
        else:
            raise TypeError(
                "translations must either be a dictionary or a Translaton object."
            )

    @property
    def talent_paths(self) -> typing.List[typing.List[SoulBindTalent]]:
        max_tier = 7

        paths: typing.List[typing.List[SoulBindTalent]] = [[]]

        for i in range(max_tier + 1):
            talents = [talent for talent in self.soul_bind_talents if talent.tier == i]

            new_paths = []
            # create a deepcopy for each
            for talent in talents:
                if talent.parent_id == 0:
                    for path in paths:
                        new_paths.append(copy.deepcopy(path) + [talent])
                else:
                    for path in paths:
                        if path[-1].id == talent.parent_id:
                            new_paths.append(copy.deepcopy(path) + [talent])
            paths = new_paths

        return paths


def _load_soul_binds() -> typing.List[SoulBind]:
    with pkg_resources.resource_stream(
        __name__, "/".join(("data_files", "soul_binds.json"))
    ) as f:
        loaded_soul_binds = json.load(f)

    def create_talent(dict, soulbind_id) -> SoulBindTalent:
        return SoulBindTalent(
            id=dict["id"],
            spell_id=dict["spell_id"],
            tier=dict["tier"],
            order=dict["ui_order"],
            parent_id=dict["id_garr_talent_prereq"],
            soulbind_id=soulbind_id,
            type=dict["conduit_type"],
            translations=_get_translations(dict),
            full_name=dict["name"],
            simc_name=dict["name"].lower().replace(" ", "_"),
        )

    soul_binds = []
    for soul_bind in loaded_soul_binds:
        soul_binds.append(
            SoulBind(
                id=soul_bind["id"],
                covenant=get_covenant(id=soul_bind["id_covenant"]),
                translations=_get_translations(soul_bind),
                full_name=soul_bind["name"],
                simc_name=soul_bind["name"].lower().replace(" ", "_"),
                soul_bind_talents=[
                    create_talent(talent, soul_bind["id"])
                    for talent in soul_bind["talents"]
                ],
            )
        )
    return soul_binds


SOULBINDS = _load_soul_binds()


def get_soul_bind(*, id: int = None, name: str = None) -> SoulBind:
    if not id and not name:
        raise ValueError("Function requires either 'id' or 'name'.")

    for soul_bind in SOULBINDS:
        if name and name == soul_bind.full_name or id and id == soul_bind.id:
            return soul_bind
    raise ValueError(f"No Soul Bind with id '{id}' or name '{name}' found.")



import unittest

from simc_support.game_data.SoulBind import SOULBINDS
from simc_support.game_data.SoulBind import SoulBind
from simc_support.game_data.SoulBind import get_soul_bind


class TestSoulBind(unittest.TestCase):
    def test_non_empty(self):
        self.assertTrue(len(SOULBINDS) > 0)

    def test_known_count(self):
        self.assertTrue(len(SOULBINDS) == 4 * 3)

    def test_path_existence(self):
        for soul_bind in SOULBINDS:
            with self.subTest(soul_bind=soul_bind):
                self.assertTrue(len(soul_bind.talent_paths) > 0)

    def test_path_length(self):
        for soul_bind in SOULBINDS:
            with self.subTest(soul_bind=soul_bind):
                for path in soul_bind.talent_paths:
                    with self.subTest(path=path):
                        self.assertEqual(len(path), 8)


class TestGetSoulBind(unittest.TestCase):
    def setUp(self) -> None:
        self.mikanikos = "Forgelite Prime Mikanikos"

    def test_return_type(self):
        self.assertTrue(isinstance(get_soul_bind(name=self.mikanikos), SoulBind))

    def test_return_value(self):
        self.assertTrue(get_soul_bind(name=self.mikanikos).full_name == self.mikanikos)



from setuptools import find_packages, setup


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="simc_support",
    author="Bloodmallet(EU)",
    author_email="bloodmalleteu@gmail.com",
    description="Data to support simulations for World of Warcraft with SimulationCraft.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/Bloodmallet/simc_support",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    package_data={
        "": [
            "*.md",
        ],
        "simc_support": [
            "game_data/data_files/*.json",
        ],
    },
    python_requires=">=3.6",
    license="GNU GENERAL PUBLIC LICENSE",
    version_config={
        "template": "{tag}",
        "dev_template": "{tag}.dev1+{branch}.{ccount}.{sha}",
        "dirty_template": "{tag}.dev1+{branch}.{ccount}.{sha}.dirty",
        # schema: <wow_version, e.g. 9.0.1>.<own_version_per_wow_version>
        "starting_version": "0.0.0.1",
        "version_callback": None,
        "version_file": None,
        "count_commits_from_version_file": False,
        "branch_formatter": None,
    },
    setup_requires=["setuptools-git-versioning"],
)



import unittest

from simc_support.game_data.Stat import Stat
from simc_support.game_data.Trinket import get_versatility_trinket
from simc_support.game_data.Trinket import Trinket
from simc_support.game_data.Trinket import TRINKETS


class TestTrinkets(unittest.TestCase):
    def test_type(self):
        self.assertTrue(isinstance(TRINKETS, list))

    def test_non_emptyness(self):
        self.assertTrue(len(TRINKETS) > 0)

    def test_for_each_primary_stat(self):
        """At least one trinket for each primary stat is available."""
        for stat in Stat:
            with self.subTest(stat=stat):
                found = False
                for trinket in TRINKETS:
                    if stat in trinket.stats:
                        found = True
                        break
                self.assertTrue(found)

    def test_for_on_use(self):
        found = False
        for trinket in TRINKETS:
            if trinket.on_use:
                found = True
                break
        self.assertTrue(found)

    def test_for_not_on_use(self):
        found = False
        for trinket in TRINKETS:
            if not trinket.on_use:
                found = True
                break
        self.assertTrue(found)

    def test_for_itemlevel_presence(self):
        """Test fails if a trinket doesn't have any itemlevels associated with it."""
        self.assertTrue(
            len(list([1 for trinket in TRINKETS if len(trinket.itemlevels) == 0])) == 0
        )


class TestGetVersatilityTrinket(unittest.TestCase):
    def test_type(self):
        self.assertTrue(isinstance(get_versatility_trinket(Stat.STRENGTH), Trinket))

    def test_stats(self):
        for stat in Stat:
            with self.subTest(stat=stat):
                trinket = get_versatility_trinket(stat)
                self.assertTrue(stat in trinket.stats)



import unittest

from simc_support.game_data.Covenant import COVENANTS
from simc_support.game_data.Covenant import Covenant
from simc_support.game_data.Covenant import get_covenant


class TestCovenants(unittest.TestCase):
    def test_non_empty(self):
        self.assertTrue(len(COVENANTS) > 0)

    def test_known_count(self):
        self.assertTrue(len(COVENANTS) == 4)


class TestGetCovenant(unittest.TestCase):
    def setUp(self) -> None:
        self.kyrian = "Kyrian"
        self.night_fae = "Night Fae"

    def test_return_type(self):
        self.assertTrue(isinstance(get_covenant(name=self.kyrian), Covenant))
        self.assertTrue(isinstance(get_covenant(name=self.night_fae), Covenant))

    def test_return_value(self):
        self.assertTrue(get_covenant(name=self.night_fae).full_name == self.night_fae)
